import sys
import subprocess

from bcipy.gui.main import BCIGui, app, AlertMessageType, AlertMessageResponse, ScrollableFrame, LineItems

from bcipy.config import BCIPY_ROOT, DEFAULT_EXPERIMENT_PATH, EXPERIMENT_FILENAME
from bcipy.helpers.load import load_experiments, load_fields
from bcipy.helpers.save import save_experiment_data


class ExperimentRegistry(BCIGui):
    """Experiment Registry.

    User interface for creating new experiments for use in BCInterface.py.
    """

    padding = 100
    btn_height = 50
    default_text = '...'
    alert_title = 'Experiment Registry Alert'
    alert_timeout = 10
    experiment_fields = []

    def __init__(self, *args, **kwargs):
        super(ExperimentRegistry, self).__init__(*args, **kwargs)

        # Structure of an experiment:
        #   { name: { fields : {name: '', required: bool, anonymize: bool}, summary: '' } }
        self.update_experiment_data()

        # These are set in the build_inputs and represent text inputs from the user
        self.name_input = None
        self.summary_input = None
        self.field_input = None
        self.panel = None
        self.line_items = None

        # fields is for display of registered fields
        self.fields = []
        self.registered_fields = load_fields()
        self.name = None
        self.summary = None

        # for registered fields
        self.build_scroll_area()

        self.show_gui()
        self.update_field_list()

    def build_scroll_area(self) -> None:
        """Build Scroll Area.

        Appends a scrollable area at the bottom of the UI for management of registered fields via LineItems.
        """
        line_widget = LineItems([], self.width)
        self.panel = ScrollableFrame(200, self.width, background_color='white', widget=line_widget)
        self.add_widget(self.panel)

    def refresh_field_panel(self) -> None:
        """Refresh Field Panel.

        Reconstruct the line items from the registered fields and refresh the scrollable panel of registered fields.
        """
        self.build_line_items_from_fields()
        self.panel.refresh(self.line_items)

    def toggle_required_field(self) -> None:
        """Toggle Required Field.

        *Button Action*

        Using the field_name retrieved from the button (get_id), find the field in self.experiment_fields and toggle
            the required field ('true' or 'false').
        """
        field_name = self.window.sender().get_id()
        for field in self.experiment_fields:
            if field_name in field:
                required = field[field_name]['required']
                if required == 'false':
                    field[field_name]['required'] = 'true'
                else:
                    field[field_name]['required'] = 'false'
        self.refresh_field_panel()

    def toggle_anonymize_field(self) -> None:
        """Toggle Anonymize Field.

        *Button Action*

        Using the field_name retrieved from the button (get_id), find the field in self.experiment_fields and toggle
            the anonymize field ('true' or 'false').
        """
        field_name = self.window.sender().get_id()
        for field in self.experiment_fields:
            if field_name in field:
                anonymize = field[field_name]['anonymize']
                if anonymize == 'false':
                    field[field_name]['anonymize'] = 'true'
                else:
                    field[field_name]['anonymize'] = 'false'
        self.refresh_field_panel()

    def remove_field(self) -> None:
        """Remove Field.

        *Button Action*

        Using the field_name retrieved from the button (get_id), find the field in self.experiment_fields and remove it
            from the list.
        """
        field_name = self.window.sender().get_id()

        idx = 0
        remove = None
        for field in self.experiment_fields:
            if field_name in field:
                remove = idx
                break
            idx += 1

        self.experiment_fields.pop(remove)
        self.refresh_field_panel()

    def build_line_items_from_fields(self) -> None:
        """Build Line Items From Fields.

        Loop over the registered experiment fields and create LineItems, which can by used to toggle the required
            field, anonymization, or remove as a registered experiment field.
        """
        items = []
        for field in self.experiment_fields:
            # experiment fields is a list of dicts, here we loop over the dict to get
            # the field_name, anonymization, and requirement
            for field_name, required in field.items():

                # Set the button text and colors, based on the requirement and anonymization
                if required['required'] == 'false':
                    required_button_label = 'Optional'
                    required_button_color = 'black'
                    required_button_text_color = 'white'
                else:
                    required_button_label = 'Required'
                    required_button_color = 'green'
                    required_button_text_color = 'white'

                if required['anonymize'] == 'false':
                    anon_button_label = 'Onymous'
                    anon_button_color = 'black'
                    anon_button_text_color = 'white'
                else:
                    anon_button_label = 'Anonymous'
                    anon_button_color = 'green'
                    anon_button_text_color = 'white'

                # Construct the item to turn into a LineItem, we set the id as field_name to use later via the action
                item = {
                    field_name: {
                        required_button_label: {
                            'action': self.toggle_required_field,
                            'color': required_button_color,
                            'textColor': required_button_text_color,
                            'id': field_name
                        },
                        anon_button_label: {
                            'action': self.toggle_anonymize_field,
                            'color': anon_button_color,
                            'textColor': anon_button_text_color,
                            'id': field_name
                        },
                        'Remove': {
                            'action': self.remove_field,
                            'color': 'red',
                            'textColor': 'white',
                            'id': field_name
                        }
                    }
                }
                items.append(item)

        # finally, set the new line items for rendering
        self.line_items = LineItems(items, self.width)

    def build_text(self) -> None:
        """Build Text.

        Build all static text needed for the UI.
        Positions are relative to the height / width of the UI defined in start_app.
        """
        text_x = 25
        text_y = 70
        font_size = 18
        self.add_static_textbox(
            text='Create BciPy Experiment',
            position=[self.width / 2 - self.padding - 50, 0],
            size=[300, 100],
            background_color='black',
            text_color='white',
            font_size=22
        )
        self.add_static_textbox(
            text='Name',
            position=[text_x, text_y],
            size=[200, 50],
            background_color='black',
            text_color='white',
            font_size=font_size)
        text_y += self.padding
        self.add_static_textbox(
            text='Summary',
            position=[text_x, text_y],
            size=[300, 50],
            background_color='black',
            text_color='white',
            font_size=font_size)
        text_y += self.padding
        self.add_static_textbox(
            text='Fields',
            position=[text_x, text_y],
            size=[300, 50],
            background_color='black',
            text_color='white',
            font_size=font_size)
        text_y += self.padding + 45
        self.add_static_textbox(
            text='Registered fields *click to toggle required field*',
            position=[text_x, text_y],
            size=[300, 50],
            background_color='black',
            text_color='white',
            font_size=14)

    def build_inputs(self) -> None:
        """Build Inputs.

        Build all text entry inputs for the UI.
        """
        input_x = 50
        input_y = 120
        input_size = [280, 25]
        self.name_input = self.add_combobox(
            position=[input_x, input_y],
            size=input_size,
            items=[self.default_text],
            editable=True,
            background_color='white',
            text_color='black')

        input_y += self.padding
        self.summary_input = self.add_combobox(
            position=[input_x, input_y],
            size=input_size,
            items=[self.default_text],
            editable=True,
            background_color='white',
            text_color='black')

        input_y += self.padding
        self.field_input = self.add_combobox(
            position=[input_x, input_y],
            size=input_size,
            items=self.fields,
            editable=False,
            background_color='white',
            text_color='black')

    def build_buttons(self):
        """Build Buttons.

        Build all buttons necessary for the UI. Define their action on click using the named argument action.
        """
        btn_create_x = self.width - self.padding - 10
        btn_create_y = self.height - self.padding - 200
        size = 150
        self.add_button(
            message='Create Experiment', position=[btn_create_x - (size / 2), btn_create_y],
            size=[size, self.btn_height],
            background_color='green',
            action=self.create_experiment,
            text_color='white')

        btn_field_x = (self.width / 2) + 150
        btn_field_y = 310
        # create field
        self.add_button(
            message='+',
            position=[btn_field_x, btn_field_y],
            size=[40, self.btn_height - 10],
            background_color='green',
            action=self.create_field,
            text_color='white'
        )

        # add field
        self.add_button(
            message='Register',
            position=[btn_field_x - 75, btn_field_y],
            size=[60, self.btn_height - 10],
            background_color='grey',
            action=self.add_field,
            text_color='white'
        )

    def create_experiment(self) -> None:
        """Create Experiment.

        After inputing all required fields, verified by check_input, add it to the experiment list and save it.
        """
        if self.check_input():
            self.add_experiment()
            self.save_experiments()
            self.throw_alert_message(
                title=self.alert_title,
                message='Experiment saved successfully! Please exit window or create another experiment!',
                message_type=AlertMessageType.INFO,
                message_response=AlertMessageResponse.OTE,
                message_timeout=self.alert_timeout
            )
            self.update_experiment_data()

    def update_experiment_data(self):
        """Update Experiment Data.

        Fetches the experiments and extracts the registered names.
        """
        self.experiments = load_experiments()
        self.experiment_names = self.experiments.keys()

    def add_experiment(self) -> None:
        """Add Experiment:

        Add a new experiment to the dict of experiments. It follows the format:
             { name: { fields : {name: '', required: bool, anonymize: bool}, summary: '' } }
        """
        self.experiments[self.name] = {
            'fields': self.experiment_fields,
            'summary': self.summary
        }

    def save_experiments(self) -> None:
        """Save Experiment.

        Save the experiments registered to the correct path as pulled from system_utils.
        """
        # add fields to the experiment
        save_experiment_data(self.experiments, self.registered_fields, DEFAULT_EXPERIMENT_PATH, EXPERIMENT_FILENAME)

    def create_field(self) -> None:
        """Create Field.

        Launch to FieldRegistry to create a new field for experiments.
        """
        subprocess.call(
            f'python {BCIPY_ROOT}/gui/experiments/FieldRegistry.py',
            shell=True)

        self.update_field_list()

    def add_field(self) -> None:
        """Add Field.

        Functionality to add fields to the newly created experiment. It will ensure no duplicates are addded.
        """
        # get the current field value and compute a list of field names already added
        field = self.field_input.currentText()
        registered_fields = [name for field in self.experiment_fields for name in field.keys()]

        # if the field selected is already registered throw an alert to the user
        if field in registered_fields:
            return self.throw_alert_message(
                title=self.alert_title,
                message=f'{field} already registered with this experiment!',
                message_type=AlertMessageType.INFO,
                message_response=AlertMessageResponse.OTE,
                message_timeout=self.alert_timeout,
            )

        # else add the field!
        self.experiment_fields.append(
            {
                field:
                    {
                        'required': 'false',
                        'anonymize': 'true'
                    }
            }
        )

        self.refresh_field_panel()

    def update_field_list(self) -> None:
        """Updates the field_input combo box with a list of fields. """

        self.field_input.clear()
        self.field_input.addItem(ExperimentRegistry.default_text)
        self.registered_fields = load_fields()
        self.fields = [item for item in self.registered_fields]
        self.field_input.addItems(self.fields)

    def build_assets(self) -> None:
        """Build Assets.

        Define the assets to build in the UI.
        """
        self.build_inputs()
        self.build_text()
        self.build_buttons()

    def check_input(self) -> bool:
        """Check Input.

        Checks to make sure user has input all required fields. Currently, only name and summary are required.
        """
        self.name = self.name_input.currentText()
        self.summary = self.summary_input.currentText()
        try:
            if self.name == ExperimentRegistry.default_text:
                self.throw_alert_message(
                    title=self.alert_title,
                    message='Please add an Experiment Name!',
                    message_type=AlertMessageType.WARN,
                    message_response=AlertMessageResponse.OTE,
                    message_timeout=self.alert_timeout)
                return False
            if self.name in self.experiment_names:
                self.throw_alert_message(
                    title=self.alert_title,
                    message=(
                        'Experiment name already registered. \n'
                        'Please use a unique Experiment name! \n'
                        f'Registed names: {self.experiment_names}'
                    ),
                    message_type=AlertMessageType.WARN,
                    message_response=AlertMessageResponse.OTE,
                    message_timeout=self.alert_timeout)
                return False
            if self.summary == ExperimentRegistry.default_text:
                self.throw_alert_message(
                    title=self.alert_title,
                    message='Please add an Experiment Summary!',
                    message_type=AlertMessageType.WARN,
                    message_response=AlertMessageResponse.OTE,
                    message_timeout=self.alert_timeout)
                return False
        except Exception as e:
            self.throw_alert_message(
                title=self.alert_title,
                message=f'Error, {e}',
                message_type=AlertMessageType.CRIT,
                message_response=AlertMessageResponse.OTE,
                message_timeout=self.alert_timeout)
            return False
        return True


def start_app() -> None:
    """Start Experiment Registry."""
    bcipy_gui = app(sys.argv)
    ex = ExperimentRegistry(
        title='Experiment Registry',
        height=700,
        width=600,
        background_color='black')

    sys.exit(bcipy_gui.exec_())


if __name__ == '__main__':
    start_app()



"""Display for presenting stimuli in a grid."""
from typing import Dict, List, Optional, Tuple, NamedTuple
import logging

from psychopy import visual, core

from bcipy.display import Display, StimuliProperties, InformationProperties, BCIPY_LOGO_PATH
from bcipy.helpers.stimuli import resize_image
from bcipy.helpers.triggers import _calibration_trigger
from bcipy.helpers.symbols import alphabet
from bcipy.display.components.task_bar import TaskBar


class SymbolDuration(NamedTuple):
    """Represents a symbol and its associated duration to display"""
    symbol: str
    duration: float
    color: str = 'white'


class MatrixDisplay(Display):
    """Matrix Display Object for Inquiry Presentation.

    Animates display objects in matrix grid common to any Matrix task.

    NOTE: The following are recommended parameter values for matrix experiments:

    time_fixation: 2
    stim_pos_x: -0.6
    stim_pos_y: 0.4
    stim_height: 0.1
    """

    def __init__(
            self,
            window: visual.Window,
            experiment_clock: core.Clock,
            stimuli: StimuliProperties,
            task_bar: TaskBar,
            info: InformationProperties,
            trigger_type: str = 'text',
            symbol_set: Optional[List[str]] = None,
            should_prompt_target: bool = True):
        """Initialize Matrix display parameters and objects.

        PARAMETERS:
        ----------
        # Experiment
        window(visual.Window): PsychoPy Window
        experiment_clock(core.Clock): Clock used to timestamp display onsets

        # Stimuli
        stimuli(StimuliProperties): attributes used for inquiries

        # Task
        task_bar(TaskBar): used for task tracking. Ex. 1/100

        # Info
        info(InformationProperties): attributes to display informational stimuli alongside task and inquiry stimuli.

        trigger_type(str) default 'image': defines the calibration trigger type for the display at the beginning of any
            task. This will be used to reconcile timing differences between acquisition and the display.
        symbol_set default = none : subset of stimuli to be highlighted during an inquiry
        should_prompt_target(bool): when True prompts for the target symbol. Assumes that this is
            the first symbol of each inquiry. For example: [target, fixation, *stim].
        """
        self.window = window

        self.logger = logging.getLogger(__name__)

        self.stimuli_inquiry = []
        self.stimuli_timing = []
        self.stimuli_colors = []
        self.stimuli_font = stimuli.stim_font

        assert stimuli.is_txt_stim, "Matrix display is a text only display"

        self.symbol_set = symbol_set or alphabet()

        # Set position and parameters for grid of alphabet
        self.position = stimuli.stim_pos
        self.grid_stimuli_height = .17
        self.position_increment = self.grid_stimuli_height + .05
        self.max_grid_width = 0.7

        self.grid_color = 'white'
        self.start_opacity = 0.15
        self.highlight_opacity = 0.95
        self.full_grid_opacity = 0.95

        # Trigger handling
        self.first_run = True
        self.first_stim_time = None
        self.trigger_type = trigger_type
        self._timing = []

        self.experiment_clock = experiment_clock

        self.task_bar = task_bar
        self.info_text = info.build_info_text(window)

        self.stim_registry = self.build_grid()
        self.should_prompt_target = should_prompt_target

    def schedule_to(self, stimuli: list, timing: list, colors: list) -> None:
        """Schedule stimuli elements (works as a buffer).

        Args:
                stimuli(list[string]): list of stimuli text / name
                timing(list[float]): list of timings of stimuli
                colors(list[string]): list of colors
        """
        assert len(stimuli) == len(
            timing), "each stimuli must have a timing value"
        self.stimuli_inquiry = stimuli
        self.stimuli_timing = timing
        if colors:
            assert len(stimuli) == len(colors), "each stimuli must have a color"
            self.stimuli_colors = colors
        else:
            self.stimuli_colors = [self.grid_color] * len(stimuli)

    def symbol_durations(self) -> List[SymbolDuration]:
        """Symbols associated with their duration for the currently configured
        stimuli_inquiry."""
        return [
            SymbolDuration(*sti)
            for sti in zip(self.stimuli_inquiry, self.stimuli_timing, self.stimuli_colors)
        ]

    def add_timing(self, stimuli: str):
        """Add a new timing entry using the stimuli as a label.

        Useful as a callback function to register a marker at the time it is
        first displayed."""
        self._timing.append([stimuli, self.experiment_clock.getTime()])

    def reset_timing(self):
        """Reset the trigger timing."""
        self._timing = []

    def do_inquiry(self) -> List[float]:
        """Animates an inquiry of stimuli and returns a list of stimuli trigger timing."""
        self.reset_timing()
        symbol_durations = self.symbol_durations()

        if self.first_run:
            self._trigger_pulse()

        if self.should_prompt_target:
            [target, fixation, *stim] = symbol_durations
            self.prompt_target(target)
        else:
            [fixation, *stim] = symbol_durations

        self.animate_scp(fixation, stim)

        return self._timing

    def build_grid(self) -> Dict[str, visual.TextStim]:
        """Build grid.

        Builds a 7x4 matrix of stimuli.
        """
        grid = {}
        pos = self.position
        for sym in self.symbol_set:
            grid[sym] = visual.TextStim(win=self.window,
                                        text=sym,
                                        color=self.grid_color,
                                        opacity=self.start_opacity,
                                        pos=pos,
                                        height=self.grid_stimuli_height)
            pos = self.increment_position(pos)
        return grid

    def increment_position(self, pos: Tuple[float, float]) -> Tuple[float, float]:
        """Computes the position of the next symbol in the matrix."""
        x_coordinate, y_coordinate = pos
        x_coordinate += self.position_increment
        if x_coordinate >= self.max_grid_width:
            y_coordinate -= self.position_increment
            x_coordinate = self.position[0]
        return (x_coordinate, y_coordinate)

    def draw_grid(self,
                  opacity: float = 1,
                  color: Optional[str] = 'white',
                  highlight: Optional[str] = None,
                  highlight_color: Optional[str] = None):
        """Draw the grid.

        Parameters
        ----------
            opacity - opacity for each item in the matrix
            color - optional color for each item in the matrix
            highlight - optional stim label for the item to be highlighted
                (rendered using the highlight_opacity).
            highlight_color - optional color to use for rendering the
              highlighted stim.
        """
        for symbol, stim in self.stim_registry.items():
            stim.setOpacity(self.highlight_opacity if highlight ==
                            symbol else opacity)
            stim.setColor(highlight_color if highlight_color and
                          highlight == symbol else color)
            stim.draw()

    def prompt_target(self, target: SymbolDuration) -> float:
        """Present the target for the configured length of time. Records the
        stimuli timing information.

        Parameters
        ----------
            target - (symbol, duration) tuple
        """
        # register any timing and marker callbacks
        self.window.callOnFlip(self.add_timing, target.symbol)
        self.draw(grid_opacity=self.start_opacity,
                  duration=target.duration,
                  highlight=target.symbol,
                  highlight_color=target.color)

    def draw(self,
             grid_opacity: float,
             grid_color: Optional[str] = None,
             duration: Optional[float] = None,
             highlight: Optional[str] = None,
             highlight_color: Optional[str] = None):
        """Draw all screen elements and flip the window.

        Parameters
        ----------
            grid_opacity - opacity value to use on all grid symbols
            grid_color - optional color to use for all grid symbols
            duration - optional seconds to wait after flipping the window.
            highlight - optional symbol to highlight in the grid.
            highlight_color - optional color to use for rendering the
              highlighted stim.
        """
        self.draw_grid(opacity=grid_opacity,
                       color=grid_color or self.grid_color,
                       highlight=highlight,
                       highlight_color=highlight_color)
        self.draw_static()
        self.window.flip()
        if duration:
            core.wait(duration)

    def animate_scp(self, fixation: SymbolDuration,
                    stimuli: List[SymbolDuration]):
        """Animate the given stimuli using single character presentation.

        Flashes each stimuli in stimuli_inquiry for their respective flash
        times and records the timing information.
        """

        # Flashing the grid at full opacity is considered fixation.
        self.window.callOnFlip(self.add_timing, fixation.symbol)
        self.draw(grid_opacity=self.full_grid_opacity,
                  grid_color=(fixation.color if self.should_prompt_target else
                              self.grid_color),
                  duration=fixation.duration / 2)
        self.draw(grid_opacity=self.start_opacity,
                  duration=fixation.duration / 2)

        for stim in stimuli:
            self.window.callOnFlip(self.add_timing, stim.symbol)
            self.draw(grid_opacity=self.start_opacity,
                      duration=stim.duration,
                      highlight=stim.symbol,
                      highlight_color=stim.color)
        self.draw(self.start_opacity)

    def wait_screen(self, message: str, message_color: str) -> None:
        """Wait Screen.

        Define what happens on the screen when a user pauses a session.
        """

        # Construct the wait message
        wait_message = visual.TextStim(win=self.window,
                                       font=self.stimuli_font,
                                       text=message,
                                       height=.1,
                                       color=message_color,
                                       pos=(0, -.5),
                                       wrapWidth=2)

        # try adding the BciPy logo to the wait screen
        try:
            wait_logo = visual.ImageStim(
                self.window,
                image=BCIPY_LOGO_PATH,
                pos=(0, .25),
                mask=None,
                ori=0.0)
            wait_logo.size = resize_image(
                BCIPY_LOGO_PATH,
                self.window.size,
                1)
            wait_logo.draw()

        except Exception as e:
            self.logger.exception(f'Cannot load logo image from path=[{BCIPY_LOGO_PATH}]')
            raise e

        # Draw and flip the screen.
        wait_message.draw()
        self.window.flip()

    def draw_static(self) -> None:
        """Draw static elements in a stimulus."""
        if self.task_bar:
            self.task_bar.draw()

        for info in self.info_text:
            info.draw()

    def update_task_bar(self, text: str = ''):
        """Update Task.

        Update any task related display items not related to the inquiry. Ex. stimuli count 1/200.

        PARAMETERS:

        text: text for task
        color_list: list of the colors for each stimuli
        pos: position of task
        """
        if self.task_bar:
            self.task_bar.update(text)

    def _trigger_pulse(self) -> None:
        """Trigger Pulse.

        This method uses a calibration trigger to determine any functional
            offsets needed for operation with this display. By setting the first_stim_time and searching for the
            same stimuli output to the marker stream, the offsets between these proceses can be reconciled at the
            beginning of an experiment. If drift is detected in your experiment, more frequent pulses and offset
            correction may be required.
        """
        calibration_time = _calibration_trigger(
            self.experiment_clock,
            trigger_type=self.trigger_type,
            display=self.window)

        # set the first stim time if not present and first_run to False
        if not self.first_stim_time:
            self.first_stim_time = calibration_time[-1]
            self.first_run = False



from bcipy.signal.evaluate.rules import HighVoltage, LowVoltage


class Evaluator:
    """Evaluator.

    Takes in raw data and tests them against given
    rules, which elicit the rejection of a inquiry when broken.
    Feeds warnings to artifact rejector as suggestions.

    Add rules given in parameters to evaluator's ruleset and set
    keys for broken_rules. One heading per rule.

    parameters (dict): dictionary of parameters from json file, given by caller
    rules (list of rule objects, defined in rules.py)
    """

    def __init__(self, parameters, high_voltage, low_voltage):
        self.rules = []

        # if high_voltage threshold is enabled, add to rules
        if high_voltage:
            self.rules.append(HighVoltage(parameters['high_voltage_value']))

        # if low_voltage threshold is enabled, add to rules
        if low_voltage:
            self.rules.append(LowVoltage(parameters['low_voltage_value']))

    def evaluate(self, data):
        """Evaluate.

        Evaluates inquiry data using selected rules from parameters file.
        """

        for rule in self.rules:

            if rule.is_broken(data):

                return False

        return True

    def __str__(self):
        rules = [str(rule) for rule in self.rules]
        return f'Evaluator with {rules}'



"""Module for recording and loading application state and user preferences."""
import json
from pathlib import Path
from typing import Any
from bcipy.config import DEFAULT_ENCODING, PREFERENCES_PATH, BCIPY_ROOT


class Pref:
    """Preference descriptor. When a class attribute is initialized as a Pref,
    values will be stored and retrieved from an 'entries' dict initialized in
    the instance.

    https://docs.python.org/3/howto/descriptor.html

    Parameters
    ----------
        default - default value assigned to the attribute.
    """

    def __init__(self, default: Any = None):
        self.default = default
        self.name = None

    def __set_name__(self, owner, name):
        """Called when the class assigns a Pref to a class attribute."""
        self.name = name

    def __get__(self, instance, owner=None):
        """Retrieve the value from the dict of entries."""
        return instance.entries.get(self.name, self.default)

    def __set__(self, instance, value):
        """Stores the given value in the entries dict keyed on the attribute
        name."""
        instance.entries[self.name] = value
        instance.save()


class Preferences:
    """User preferences persisted to disk to retain application state between
    work sessions.

    Parameters
    ----------
        filename - optional file used for persisting entries.
    """
    signal_model_directory: str = Pref()
    last_directory: str = Pref(default=str(BCIPY_ROOT))

    def __init__(self, filename: str = PREFERENCES_PATH):
        self.filename = filename
        self.entries = {}
        self.load()

    def load(self):
        """Load preference data from the persisted file."""
        if Path(self.filename).is_file():
            with open(self.filename, 'r',
                      encoding=DEFAULT_ENCODING) as json_file:
                for key, val in json.load(json_file).items():
                    self.entries[key] = val

    def save(self):
        """Write preferences to disk."""
        with open(self.filename, 'w', encoding=DEFAULT_ENCODING) as json_file:
            json.dump(self.entries, json_file, ensure_ascii=False, indent=2)

    def get(self, name: str):
        """Get preference by name"""
        return self.entries.get(name, None)

    def set(self, name: str, value: Any, persist: bool = True):
        """Set a preference and save the result.

        Parameters
        ----------
            name - name of the preference
            value - value associated with the given name
            persist - flag indicating whether to immediately save the result.
                Default is True.
        """
        self.entries[name] = value
        if persist:
            self.save()


preferences = Preferences()



"""Records LSL data streams to a data store."""
import logging
import time
from pathlib import Path
from typing import List, Tuple

from pylsl import StreamInfo, StreamInlet, resolve_streams

from bcipy.acquisition.devices import DeviceSpec
from bcipy.acquisition.protocols.lsl.lsl_connector import channel_names, check_device
from bcipy.acquisition.util import StoppableThread
from bcipy.helpers.raw_data import RawDataWriter

log = logging.getLogger(__name__)


class LslRecorder:
    """Records LSL data to a datastore. Resolves streams when started.

    Parameters:
    -----------
    - path : location to store the recordings
    - filenames : optional dict mapping device type to its raw data filename.
    Devices without an entry will use a naming convention.
    """

    def __init__(self, path: str, filenames: dict = None):
        super().__init__()
        self.path = path
        self.streams = None
        self.filenames = filenames or {}

    def start(self):
        """Start recording all streams currently on the network."""

        if not self.streams:
            log.debug("Recording data")
            # create a thread for each.
            self.streams = [
                LslRecordingThread(stream, self.path,
                                   self.filenames.get(stream.type(), None))
                for stream in resolve_streams()
            ]

            # Validate that streams have unique names for their type
            stream_names = [stream.filename for stream in self.streams]
            if len(stream_names) != len(set(stream_names)):
                raise Exception("Data stream names are not unique")
            for stream in self.streams:
                stream.start()

    def stop(self, wait: bool = False):
        """Stop recording.

        Parameters
        ----------
        - wait : if True waits for all threads to stop before returning.
        """
        for stream in self.streams:
            stream.stop()
            if wait:
                stream.join()
        self.streams = None


class LslRecordingThread(StoppableThread):
    """Records data for the given LabStreamingLayer (LSL) data stream.

    Parameters:
    ----------
    - stream : information about the stream of interest
    - directory : location to store the recording
    - filename : optional, name of the data file.
    - device_spec : optional DeviceSpec ; if provided channel labels will come
        from here.
    """

    def __init__(self,
                 stream_info: StreamInfo,
                 directory: str,
                 filename: str = None,
                 device_spec: DeviceSpec = None):
        super().__init__()
        self.stream_info = stream_info
        self.directory = directory
        self.device_spec = device_spec

        self.sample_count = 0
        # see: https://labstreaminglayer.readthedocs.io/info/faqs.html#chunk-sizes
        self.max_chunk_size = 1024

        # seconds to sleep between data pulls from LSL
        self.sleep_seconds = 0.2
        self.writer = None

        self.filename = filename if filename else self.default_filename()
        self.first_sample_time = None
        self.last_sample_time = None

    def default_filename(self):
        """Default filename to use if a name is not provided."""
        content_type = '_'.join(self.stream_info.type().split()).lower()
        name = '_'.join(self.stream_info.name().split()).lower()
        return f"{content_type}_data_{name}.csv"

    @property
    def recorded_seconds(self) -> float:
        """Total seconds of data recorded."""
        if self.first_sample_time and self.last_sample_time:
            return self.last_sample_time - self.first_sample_time
        return 0.0

    def _init_data_writer(self, stream_info: StreamInfo) -> None:
        """Initializes the raw data writer.

        Parameters:
        ----------
        - metadata : metadata about the data stream.
        """
        assert self.writer is None, "Data store has already been initialized."

        channels = channel_names(stream_info)
        # Use the device_spec channels labels if provided.
        if self.device_spec:
            check_device(self.device_spec, stream_info)
            channels = self.device_spec.channels

        path = Path(self.directory, self.filename)
        log.debug(f"Writing data to {path}")
        self.writer = RawDataWriter(
            path,
            daq_type=self.stream_info.name(),
            sample_rate=self.stream_info.nominal_srate(),
            columns=['timestamp'] + channels + ['lsl_timestamp'])
        self.writer.__enter__()

    def _cleanup(self) -> None:
        """Performs cleanup tasks."""
        assert self.writer, "Writer not initialized"
        self.writer.__exit__()
        self.writer = None

    def _write_chunk(self, data: List, timestamps: List) -> None:
        """Persists the data resulting from pulling a chunk from the inlet.

        Parameters
        ----------
            data : list of samples
            timestamps : list of timestamps
        """
        assert self.writer, "Writer not initialized"

        chunk = []
        for i, sample in enumerate(data):
            self.sample_count += 1
            chunk.append([self.sample_count] + sample + [timestamps[i]])
        self.writer.writerows(chunk)

    def _pull_chunk(self, inlet: StreamInlet) -> Tuple[int, float]:
        """Pull a chunk of data and persist. Updates first_sample_time,
        last_sample_time, and sample_count.

        Parameters
        ----------
            inlet : stream inlet from which to pull

        Returns
        -------
            number of samples pulled
        """
        # A timeout of 0.0 does not block and only gets samples immediately
        # available.
        data, timestamps = inlet.pull_chunk(timeout=0.0,
                                            max_samples=self.max_chunk_size)
        if timestamps:
            if not self.first_sample_time:
                self.first_sample_time = timestamps[0]
            self.last_sample_time = timestamps[-1]
            self._write_chunk(data, timestamps)
        return len(timestamps)

    def _reset(self) -> None:
        """Reset state"""
        self.sample_count = 0
        self.first_sample_time = None
        self.last_sample_time = None

    # @override
    def run(self):
        """Process startup. Connects to the device, reads chunks of data at the
        given interval, and persists the results. This happens continuously
        until the `stop()` method is called.
        """
        # Note that self.stream_info does not have the channel names.
        inlet = StreamInlet(self.stream_info)
        full_metadata = inlet.info()

        log.debug("Acquiring data from data stream:")
        log.debug(full_metadata.as_xml())

        self._reset()
        self._init_data_writer(full_metadata)

        # TODO: account for remote acquisition by recording remote clock offsets
        # so we can map from remote timestamp to local lsl clock for comparing
        # datasets.

        # Run loop for continuous acquisition
        while self.running():
            self._pull_chunk(inlet)
            time.sleep(self.sleep_seconds)

        # Pull any remaining samples up to the current time.
        log.debug("Pulling remaining samples")
        record_count = self._pull_chunk(inlet)
        while record_count == self.max_chunk_size:
            record_count = self._pull_chunk(inlet)

        log.info(f"Ending data stream recording for {self.stream_info.name()}")
        log.info(f"Total recorded seconds: {self.recorded_seconds}")
        log.info(f"Total recorded samples: {self.sample_count}")
        inlet.close_stream()
        self._cleanup()


def main(path: str, seconds: int = 5, debug: bool = False):
    """Function to demo the LslRecorder. Expects LSL data streams to be already
    running."""
    if debug:
        # pylint: disable=import-outside-toplevel
        from bcipy.helpers.system_utils import log_to_stdout
        log_to_stdout()
    recorder = LslRecorder(path)
    print(f"\nCollecting data for {seconds}s...")
    recorder.start()
    try:
        time.sleep(seconds)
    except KeyboardInterrupt:
        print("Keyboard Interrupt; stopping.")
    finally:
        recorder.stop()


if __name__ == '__main__':
    import argparse
    # pylint: disable=invalid-name
    parser = argparse.ArgumentParser()
    parser.add_argument('--path', default='.')
    parser.add_argument('--seconds', default=5)
    parser.add_argument('--debug', action='store_true')
    parsed_args = parser.parse_args()
    main(path=parsed_args.path,
         seconds=int(parsed_args.seconds),
         debug=parsed_args.debug)



import subprocess
import sys

from typing import List

from bcipy.config import BCIPY_ROOT, DEFAULT_PARAMETERS_PATH, STATIC_IMAGES_PATH
from bcipy.gui.main import (
    AlertMessageResponse,
    AlertMessageType,
    AlertResponse,
    app,
    BCIGui,
    contains_special_characters,
    contains_whitespaces,
    invalid_length,
)
from bcipy.helpers.load import load_json_parameters, load_experiments, copy_parameters, load_users
from bcipy.task import TaskType


class BCInterface(BCIGui):
    """BCI Interface.

    Main interface for execution of BciPy experiments and tasks. Additionally, quick access to parameter
        editing and loading, and offline analysis execution.
    """

    tasks = TaskType.list()

    default_text = '...'
    padding = 30
    btn_height = 40
    max_length = 25
    min_length = 1
    timeout = 3

    def __init__(self, *args, **kwargs):
        super(BCInterface, self).__init__(*args, **kwargs)
        self.parameter_location = DEFAULT_PARAMETERS_PATH

        self.parameters = load_json_parameters(self.parameter_location,
                                               value_cast=True)

        # These are set in the build_inputs and represent text inputs from the user
        self.user_input = None
        self.experiment_input = None
        self.task_input = None

        # These represent the current user, experiment, and task selected in the gui
        self.user = None
        self.experiment = None
        self.task = None

        # user names available in the dropdown menu
        self.users = []

        # setup a timer to prevent double clicking in gui
        self.disable = False
        self.timer.timeout.connect(self._disable_action)

        self.task_start_timeout = self.timeout
        self.button_timeout = self.timeout

        self.autoclose = False
        self.alert = True
        self.static_font_size = 24

        self.user_id_validations = [
            (invalid_length(min=self.min_length, max=self.max_length),
             f'User ID must contain between {self.min_length} and {self.max_length} alphanumeric characters.'),
            (contains_whitespaces, 'User ID cannot contain white spaces'),
            (contains_special_characters, 'User ID cannot contain special characters')
        ]

    def build_buttons(self) -> None:
        """Build Buttons.

        Build all buttons necessary for the UI. Define their action on click using the named argument action.
        """
        self.add_button(
            message='Load Parameters', position=[self.padding, 450],
            size=[110, self.btn_height], background_color='white',
            action=self.select_parameters)

        self.add_button(
            message='Edit Parameters', position=[self.padding + 115, 450],
            size=[105, self.btn_height], background_color='white',
            action=self.edit_parameters)

        btn_auc_width = 100
        btn_auc_x = self.padding + 225
        self.add_button(
            message='Calculate AUC', position=(btn_auc_x, 450),
            size=(btn_auc_width, self.btn_height), background_color='white',
            action=self.offline_analysis)

        btn_start_width = 200
        btn_start_x = self.width - (self.padding + btn_start_width)
        self.add_button(
            message='Start Experiment Session', position=[btn_start_x, 450],
            size=[btn_start_width, self.btn_height],
            background_color='green',
            action=self.start_experiment,
            text_color='white')

        self.add_button(
            message='+',
            position=[self.width - self.padding - 200, 260],
            size=[40, self.btn_height - 10],
            background_color='green',
            action=self.create_experiment,
            text_color='white'
        )

    def create_experiment(self) -> None:
        """Create Experiment.

        Launch the experiment registry which will be used to add new experiments for selection in the GUI.
        """
        if not self.action_disabled():
            subprocess.call(
                f'python {BCIPY_ROOT}/gui/experiments/ExperimentRegistry.py',
                shell=True)

            self.update_experiment_list()

    def update_user_list(self, refresh=True) -> None:
        """Updates the user_input combo box with a list of user ids based on the
        data directory configured in the current parameters."""
        # if refresh is True, then we need to clear the list and add the default text
        if refresh:
            self.user_input.clear()
            self.user_input.addItem(BCInterface.default_text)

        # load the users from the data directory and check if they have already been added to the dropdown
        users = load_users(self.parameters['data_save_loc'])
        for user in users:
            if user not in self.users:
                self.user_input.addItem(user)
                self.users.append(user)

    def update_experiment_list(self) -> None:
        """Updates the experiment_input combo box with a list of experiments based on the
        data directory configured in the current parameters."""

        self.experiment_input.clear()
        self.experiment_input.addItem(BCInterface.default_text)
        self.experiment_input.addItems(self.load_experiments())

    def update_task_list(self) -> None:
        """Updates the task_input combo box with a list of tasks ids based on what
        is available in the Task Registry"""

        self.task_input.clear()
        self.task_input.addItem(BCInterface.default_text)
        self.task_input.addItems(self.tasks)

    def build_inputs(self) -> None:
        """Build Inputs.

        Build all inputs needed for BCInterface.
        """
        input_x = 170
        input_y = 160
        self.user_input = self.add_combobox(
            position=[input_x, input_y],
            size=[280, 25],
            items=[],
            editable=True,
            background_color='white',
            text_color='black')

        self.update_user_list()

        input_y += 100
        self.experiment_input = self.add_combobox(
            position=[input_x, input_y],
            size=[280, 25],
            items=[],
            editable=False,
            background_color='white',
            text_color='black')

        self.update_experiment_list()

        input_y += 100
        self.task_input = self.add_combobox(
            position=[input_x, input_y],
            size=[280, 25],
            items=[],
            editable=False,
            background_color='white',
            text_color='black')

        self.update_task_list()

    def build_text(self) -> None:
        """Build Text.

        Build all static text needed for the UI.
        Positions are relative to the height / width of the UI defined in start_app.
        """
        self.add_static_textbox(
            text='BCInterface',
            position=[275, 0],
            size=[200, 50],
            background_color='black',
            text_color='white',
            font_size=30)

        text_x = 145
        self.add_static_textbox(
            text='User',
            position=[text_x, 105],
            size=[200, 50],
            background_color='black',
            text_color='white',
            font_size=self.static_font_size)
        self.add_static_textbox(
            text='Experiment',
            position=[text_x, 205],
            size=[300, 50],
            background_color='black',
            text_color='white',
            font_size=self.static_font_size)
        self.add_static_textbox(
            text='Task',
            position=[text_x, 305],
            size=[300, 50],
            background_color='black',
            text_color='white',
            font_size=self.static_font_size)

    def build_images(self) -> None:
        """Build Images.

        Build add images needed for the UI. In this case, the OHSU and NEU logos.
        """
        self.add_image(
            path=f'{STATIC_IMAGES_PATH}/gui/ohsu.png', position=[self.padding, 0], size=100)
        self.add_image(
            path=f'{STATIC_IMAGES_PATH}/gui/neu.png', position=[self.width - self.padding - 110, 0], size=100)

    def build_assets(self) -> None:
        """Build Assets.

        Define the assets to build in the UI.
        """
        self.build_buttons()
        self.build_inputs()
        self.build_text()
        self.build_images()

    def set_parameter_location(self, path: str) -> None:
        """Sets the parameter_location to the given path. Reloads the parameters
        and updates any GUI widgets that are populated based on these params."""
        self.parameter_location = path
        self.parameters = load_json_parameters(self.parameter_location,
                                               value_cast=True)
        self.update_user_list(refresh=False)

    def select_parameters(self) -> None:
        """Select Parameters.

        Opens a dialog to select the parameters.json configuration to use.
        """

        response = self.get_filename_dialog(message='Select parameters file',
                                            file_type='JSON files (*.json)')
        if response:
            self.set_parameter_location(response)
            # If outdated, prompt to merge with the current defaults
            default_parameters = load_json_parameters(DEFAULT_PARAMETERS_PATH,
                                                      value_cast=True)
            if self.parameters.add_missing_items(default_parameters):
                save_response = self.throw_alert_message(
                    title='BciPy Alert',
                    message='The selected parameters file is out of date.'
                            'Would you like to update it with the latest options?',
                    message_type=AlertMessageType.INFO,
                    message_response=AlertMessageResponse.OCE)

                if save_response == AlertResponse.OK.value:
                    self.parameters.save()

    def edit_parameters(self) -> None:
        """Edit Parameters.

        Prompts for a parameters.json file to use. If the default parameters are selected, a copy is used.
        Note that any edits to the parameters file will not be applied to this GUI until the parameters
        are reloaded.
        """
        if not self.action_disabled():
            if self.parameter_location == DEFAULT_PARAMETERS_PATH:
                # Don't allow the user to overwrite the defaults
                response = self.throw_alert_message(
                    title='BciPy Alert',
                    message='The default parameters.json cannot be overridden. A copy will be used.',
                    message_type=AlertMessageType.INFO,
                    message_response=AlertMessageResponse.OCE)

                if response == AlertResponse.OK.value:
                    self.parameter_location = copy_parameters()
                else:
                    return None

            subprocess.call(
                f'python {BCIPY_ROOT}/gui/parameters/params_form.py -p {self.parameter_location}',
                shell=True)

    def check_input(self) -> bool:
        """Check Input.

        Checks to make sure user has input all required fields.
        """

        # Update based on current inputs
        self.user = self.user_input.currentText()
        self.experiment = self.experiment_input.currentText()
        self.task = self.task_input.currentText()

        # Check the set values are different than defaults
        try:
            if not self.check_user_id():
                return False
            if self.experiment == BCInterface.default_text:
                self.throw_alert_message(
                    title='BciPy Alert',
                    message='Please select or create an Experiment',
                    message_type=AlertMessageType.INFO,
                    message_response=AlertMessageResponse.OTE)
                return False
            if self.task == BCInterface.default_text:
                self.throw_alert_message(
                    title='BciPy Alert',
                    message='Please select a Task',
                    message_type=AlertMessageType.INFO,
                    message_response=AlertMessageResponse.OTE)
                return False
        except Exception as e:
            self.throw_alert_message(
                title='BciPy Alert',
                message=f'Error, {e}',
                message_type=AlertMessageType.CRIT,
                message_response=AlertMessageResponse.OTE)
            return False
        return True

    def check_user_id(self) -> bool:
        """Check User ID

        User ID must meet the following requirements:

        1. Maximum length of self.max_length alphanumeric characters
        2. Minimum length of at least self.min_length alphanumeric character
        3. No special characters
        4. No spaces
        """
        # Check the user id set is different than the default text
        if self.user == BCInterface.default_text:
            self.throw_alert_message(
                title='BciPy Alert',
                message='Please input a User ID',
                message_type=AlertMessageType.INFO,
                message_response=AlertMessageResponse.OTE)
            return False
        # Loop over defined user validations and check for error conditions
        for validator in self.user_id_validations:
            (invalid, error_message) = validator
            if invalid(self.user):
                self.throw_alert_message(
                    title='BciPy Alert',
                    message=error_message,
                    message_type=AlertMessageType.INFO,
                    message_response=AlertMessageResponse.OTE
                )
                return False
        return True

    def load_experiments(self) -> List[str]:
        """Load experiments

        Loads experiments registered in the DEFAULT_EXPERIMENT_PATH.
        """
        return load_experiments().keys()

    def start_experiment(self) -> None:
        """Start Experiment Session.

        Using the inputs gathers, check for validity using the check_input method, then launch the experiment using a
            command to bcipy main and subprocess.
        """
        if self.check_input() and not self.action_disabled():
            self.throw_alert_message(
                title='BciPy Alert',
                message='Task Starting ...',
                message_type=AlertMessageType.INFO,
                message_response=AlertMessageResponse.OTE,
                message_timeout=self.task_start_timeout)
            cmd = (
                f'bcipy -e "{self.experiment}" '
                f'-u "{self.user}" -t "{self.task}" -p "{self.parameter_location}"'
            )
            if self.alert:
                cmd += ' -a'
            subprocess.Popen(cmd, shell=True)

            if self.autoclose:
                self.close()

    def offline_analysis(self) -> None:
        """Offline Analysis.

        Run offline analysis as a script in a new process.
        """
        if not self.action_disabled():
            cmd = f'python {BCIPY_ROOT}/signal/model/offline_analysis.py --alert --p "{self.parameter_location}"'
            subprocess.Popen(cmd, shell=True)

    def action_disabled(self) -> bool:
        """Action Disabled.

        Method to check whether another action can take place. If not disabled, it will allow the action and
        start a timer that will disable actions until self.timeout (seconds) has occured.

        Note: the timer is registed with the private method self._disable_action, which when self.timeout has
        been reached, resets self.disable and corresponding timeouts.
        """
        if self.disable:
            return True
        else:
            self.disable = True
            # set the update time to every 1000ms
            self.timer.start(1000)
            return False

    def _disable_action(self) -> bool:
        """Disable Action.

        A private method to register with a BCIGui.timer after setting self.button_timeout.
        """
        if self.button_timeout > 0:
            self.disable = True
            self.button_timeout -= 1
            return self.disable

        self.timer.stop()
        self.disable = False
        self.button_timeout = self.timeout
        return self.disable


def start_app() -> None:
    """Start BCIGui."""
    bcipy_gui = app(sys.argv)
    ex = BCInterface(
        title='Brain Computer Interface',
        height=550,
        width=750,
        background_color='black')

    ex.show_gui()

    sys.exit(bcipy_gui.exec_())


if __name__ == '__main__':
    start_app()



"""Functionality to simulate a switch"""
import logging
import sys

from pylsl import StreamInfo, StreamOutlet

from bcipy.acquisition.devices import DeviceSpec, IRREGULAR_RATE
from bcipy.gui.main import BCIGui, app

log = logging.getLogger(__name__)


def switch_device() -> DeviceSpec:
    """Mock DeviceSpec for a switch"""
    return DeviceSpec(name='Switch',
                      channels=['Marker'],
                      sample_rate=IRREGULAR_RATE,
                      content_type='Markers')


class Switch:
    """Mock switch which streams data over LSL at an irregular interval."""

    def __init__(self):
        super().__init__()
        self.device = switch_device()
        self.lsl_id = 'bci_demo_switch'
        info = StreamInfo(self.device.name, self.device.content_type,
                          self.device.channel_count, self.device.sample_rate,
                          self.device.data_type, self.lsl_id)
        self.outlet = StreamOutlet(info)

    def click(self, _position):
        """Click event that pushes a sample"""
        log.debug("Click!")
        self.outlet.push_sample([1.0])

    def quit(self):
        """Quit and cleanup"""
        del self.outlet
        self.outlet = None


class SwitchGui(BCIGui):
    """GUI to emulate a switch."""

    def __init__(self, switch: Switch, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.switch = switch

    def build_buttons(self) -> None:
        """Build all buttons necessary for the UI.
        """
        self.add_button(message='Click!',
                        position=[100, 75],
                        size=[50, 50],
                        background_color='white',
                        action=self.switch.click)

    def build_text(self) -> None:
        """Build all static text needed for the UI.
        Positions are relative to the height / width of the UI defined in start_app.
        """
        self.add_static_textbox(
            text='Click the button to emulate a switch hit.',
            position=[5, 0],
            size=[300, 50],
            background_color='black',
            text_color='white',
            font_size=16)


def main(switch: Switch):
    """Creates a PyQt5 GUI with a single button in the middle. Performs the
    switch action when clicked."""
    gui = app(sys.argv)
    ex = SwitchGui(switch=switch,
                   title='Demo Switch!',
                   height=200,
                   width=300,
                   background_color='black')

    ex.show_gui()
    result = gui.exec_()
    switch.quit()
    sys.exit(result)


if __name__ == '__main__':
    main(Switch())



from bcipy.display.paradigm.rsvp.display import RSVPDisplay
from bcipy.helpers.symbols import SPACE_CHAR


class CalibrationDisplay(RSVPDisplay):
    """Calibration Display."""

    def __init__(self,
                 window,
                 static_clock,
                 experiment_clock,
                 stimuli,
                 task_bar,
                 info,
                 trigger_type='image',
                 preview_inquiry=None,
                 space_char=SPACE_CHAR,
                 full_screen=False):

        super().__init__(window,
                         static_clock,
                         experiment_clock,
                         stimuli,
                         task_bar,
                         info,
                         trigger_type=trigger_type,
                         preview_inquiry=preview_inquiry,
                         space_char=space_char,
                         full_screen=full_screen)



import os

from bcipy.config import (
    DEFAULT_EXPERIMENT_PATH,
    DEFAULT_FIELD_PATH,
    EXPERIMENT_FILENAME,
    FIELD_FILENAME)
from bcipy.helpers.load import load_experiments, load_fields
from bcipy.helpers.system_utils import is_battery_powered, is_connected, is_screen_refresh_rate_low
from bcipy.helpers.exceptions import (InvalidFieldException,
                                      InvalidExperimentException,
                                      UnregisteredExperimentException,
                                      UnregisteredFieldException)
from bcipy.gui.alert import confirm


def validate_bcipy_session(parameters: dict, fake_data: bool) -> bool:
    """Check pre-conditions for a BciPy session. If any possible problems are
    detected, alert the user and prompt to continue.

    Parameters
    ----------
    parameters - configuration used to check for issues

    Returns
    -------
    True if it's okay to continue, otherwise False
    """
    possible_alerts = [(fake_data, '* Fake data is on.'),
                       (is_connected(), '* Internet is on.'),
                       (is_battery_powered(), '* Operating on battery power'),
                       (is_screen_refresh_rate_low(), '* Screen refresh rate is low (< 120 Hz)')]
    alert_messages = [
        message for (condition, message) in possible_alerts if condition
    ]
    if alert_messages:
        lines = [
            "The following conditions may affect system behavior:\n",
            *alert_messages, "\nDo you want to continue?"
        ]
        return confirm("\n".join(lines))
    return True


def validate_experiment(
        experiment_name: str,
        experiment_path: str = f'{DEFAULT_EXPERIMENT_PATH}/{EXPERIMENT_FILENAME}',
        field_path: str = f'{DEFAULT_FIELD_PATH}/{FIELD_FILENAME}'
) -> bool:
    """Validate Experiment.

    Validate the experiment is in the correct format and the fields are properly registered.
    """
    experiments = load_experiments(experiment_path)
    fields = load_fields(field_path)

    # attempt to load the experiment by name
    try:
        experiment = experiments[experiment_name]
    except KeyError:
        raise UnregisteredExperimentException(
            f'Experiment [{experiment_name}] is not registered at path [{experiment_path}]')

    # attempt to load the experiment by name
    _validate_experiment_format(experiment, experiment_name)
    _validate_experiment_fields(experiment['fields'], fields)

    return True


def _validate_experiment_fields(experiment_fields, fields):
    # loop over the experiment fields and attempt to load them by name
    for field in experiment_fields:
        # field is a dictionary with one key another dictionary as its' value
        # {'field_name': {require: bool, anonymize: bool}}
        field_name = list(field.keys())[0]
        try:
            fields[field_name]
        except KeyError:
            raise UnregisteredFieldException(f'Field [{field}] is not registered in [{fields}]')

        try:
            field[field_name]['required']
            field[field_name]['anonymize']
        except KeyError:
            raise InvalidFieldException(
                f'Experiment Field [{field}] incorrectly formatted. It should contain: required and anonymize')


def validate_field_data_written(path: str, file_name: str) -> bool:
    """Validate Field Data Written

    Validate that a field data file was written after executing the experiment field collection.
    """
    experiment_data_path = f'{path}/{file_name}'
    if os.path.isfile(experiment_data_path):
        return True
    raise InvalidFieldException(f'Experimental field data expected at path=[{experiment_data_path}] but not found.')


def validate_experiments(experiments, fields) -> bool:
    """Validate Experiments.

    Validate all experiments are in the correct format and the fields are properly registered.
    """
    for experiment_name in experiments:
        experiment = experiments[experiment_name]

        _validate_experiment_format(experiment, experiment_name)
        _validate_experiment_fields(experiment['fields'], fields)

    return True


def _validate_experiment_format(experiment, name):
    try:
        exp_summary = experiment['summary']
        assert isinstance(exp_summary, str)
        experiment_fields = experiment['fields']
        assert isinstance(experiment_fields, list)
    except KeyError:
        raise InvalidExperimentException(
            f'Experiment [{name}] is formatted incorrectly. It should contain the keys: summary and fields.')
    except AssertionError:
        raise InvalidExperimentException(
            f'Experiment [{name}] is formatted incorrectly. Unexpected type on summary(string) or fields(List[dict]).')



from itertools import cycle
from typing import List
from bcipy.task.paradigm.matrix.calibration import (
    MatrixCalibrationTask, InquirySchedule, DEFAULT_TEXT_FIXATION, Display,
    init_calibration_display_task)
from bcipy.helpers.stimuli import PhotoDiodeStimuli, jittered_timing


class MatrixTimingVerificationCalibration(MatrixCalibrationTask):
    """Matrix Timing Verification Task.

    This task is used for verifying display timing by alternating solid and empty boxes. These
        stimuli can be used with a photodiode to ensure accurate presentations.

    Input:
        win (PsychoPy Display Object)
        daq (Data Acquisition Object)
        parameters (Dictionary)
        file_save (String)

    Output:
        file_save (String)
    """
    TASK_NAME = 'Matrix Timing Verification Task'

    def init_display(self) -> Display:
        """Initialize the display"""
        display = init_calibration_display_task(
            self.parameters, self.window, self.experiment_clock,
            symbols_with_photodiode_stim(self.symbol_set))
        display.start_opacity = 0.0
        return display

    def generate_stimuli(self) -> InquirySchedule:
        """Generates the inquiries to be presented.
        """
        samples, times, colors = [], [], []
        [time_target, time_fixation, time_stim] = self.timing
        stimuli = cycle(PhotoDiodeStimuli.list())
        stim_timing = jittered_timing(time_stim, self.jitter, self.stim_length)

        # advance iterator to start on the solid stim.
        next(stimuli)

        # alternate between solid and empty boxes
        inq_stim = [PhotoDiodeStimuli.SOLID.value, DEFAULT_TEXT_FIXATION
                    ] + [next(stimuli) for _ in range(self.stim_length)]
        inq_times = [time_target, time_fixation
                     ] + stim_timing
        inq_colors = [self.color[0]] * (self.stim_length + 2)

        for _ in range(self.stim_number):
            samples.append(inq_stim)
            times.append(inq_times)
            colors.append(inq_colors)

        return InquirySchedule(samples, times, colors)


def symbols_with_photodiode_stim(symbols: List[str]) -> List[str]:
    """Stim symbols with the central letters swapped out for Photodiode stim.
    """
    mid = int(len(symbols) / 2)

    def sym_at_index(sym, index) -> str:
        if index == mid:
            return PhotoDiodeStimuli.SOLID.value
        if index == mid + 1:
            return PhotoDiodeStimuli.EMPTY.value
        return sym

    return [sym_at_index(sym, i) for i, sym in enumerate(symbols)]



"""Functionality for loading and querying configuration for supported hardware
devices."""
import json
import logging
from pathlib import Path
from typing import Dict, List, NamedTuple, Union

from bcipy.config import DEFAULT_ENCODING, DEVICE_SPEC_PATH

IRREGULAR_RATE = 0.0
DEFAULT_CONFIG = DEVICE_SPEC_PATH
_SUPPORTED_DEVICES = {}
# see https://labstreaminglayer.readthedocs.io/projects/liblsl/ref/enums.html
SUPPORTED_DATA_TYPES = [
    'float32', 'double64', 'string', 'int32', 'int16', 'int8'
]
DEFAULT_DEVICE_TYPE = 'EEG'

log = logging.getLogger(__name__)


class ChannelSpec(NamedTuple):
    """Represents metadata about a channel."""
    name: str  # Label in the LSL metadata
    label: str  # Label used within BciPy (raw_data, etc.)
    type: str = None
    units: str = None

    def __repr__(self):
        fields = ['name', 'label', 'type', 'units']
        items = [(field, self.__getattribute__(field)) for field in fields]
        props = [f"{key}='{val}'" for key, val in items if val]
        return f"ChannelSpec({', '.join(props)})"


def channel_spec(channel: Union[str, dict, ChannelSpec]) -> ChannelSpec:
    """Creates a ChannelSpec from the given channel.

    Parameters
    ----------
        channel - acquisition channel information specified as either just the
          label or with additional data represented by a dict or ChannelSpec.
    """
    if isinstance(channel, str):
        return ChannelSpec(name=channel, label=channel)
    if isinstance(channel, ChannelSpec):
        return channel
    if isinstance(channel, dict):
        return ChannelSpec(**channel)
    raise Exception("Unexpected channel type")


class DeviceSpec:
    """Specification for a hardware device used in data acquisition.

    Parameters
    ----------
        name - device short name; ex. DSI-24
        channels - list of data collection channels; devices must have at least
            one channel. Channels may be provided as a list of names or list of
            ChannelSpecs.
        sample_rate - sample frequency in Hz.
        content_type - type of device; likely one of ['EEG', 'MoCap', 'Gaze',
            'Audio', 'Markers']; see https://github.com/sccn/xdf/wiki/Meta-Data.
        description - device description
            ex. 'Wearable Sensing DSI-24 dry electrode EEG headset'
        data_type - data format of a channel; all channels must have the same type;
            see https://labstreaminglayer.readthedocs.io/projects/liblsl/ref/enums.html
        excluded_from_analysis - list of channels (label) to exclude from analysis.
    """

    def __init__(self,
                 name: str,
                 channels: Union[List[str], List[ChannelSpec], List[dict]],
                 sample_rate: float,
                 content_type: str = DEFAULT_DEVICE_TYPE,
                 description: str = None,
                 excluded_from_analysis: List[str] = None,
                 data_type='float32'):

        assert sample_rate >= 0, "Sample rate can't be negative."
        assert data_type in SUPPORTED_DATA_TYPES

        self.name = name
        self.channel_specs = [channel_spec(ch) for ch in channels]
        self.sample_rate = sample_rate
        self.content_type = content_type
        self.description = description or name
        self.data_type = data_type
        self.excluded_from_analysis = excluded_from_analysis or []
        self._validate_excluded_channels()

    @property
    def channel_count(self) -> int:
        """Number of channels"""
        return len(self.channel_specs)

    @property
    def channels(self) -> List[str]:
        """List of channel labels. These may be customized for BciPy."""
        return [ch.label for ch in self.channel_specs]

    @property
    def channel_names(self) -> List[str]:
        """List of channel names from the device."""
        return [ch.name for ch in self.channel_specs]

    @property
    def analysis_channels(self) -> List[str]:
        """List of channels used for analysis by the signal module.
        Parameters:
        -----------
            exclude_trg - indicates whether or not to exclude a TRG channel if present.
        """

        return list(
            filter(lambda channel: channel not in self.excluded_from_analysis,
                   self.channels))

    def to_dict(self) -> dict:
        """Converts the DeviceSpec to a dict."""
        return {
            'name': self.name,
            'content_type': self.content_type,
            'channels': [ch._asdict() for ch in self.channel_specs],
            'sample_rate': self.sample_rate,
            'description': self.description,
            'excluded_from_analysis': self.excluded_from_analysis
        }

    def __str__(self):
        """Custom str representation."""
        names = [
            'name', 'content_type', 'channels', 'sample_rate', 'description'
        ]

        def quoted_value(name):
            value = self.__getattribute__(name)
            return f"'{value}'" if isinstance(value, str) else value

        props = [f"{name}={quoted_value(name)}" for name in names]
        return f"DeviceSpec({', '.join(props)})"

    def _validate_excluded_channels(self):
        """Warn if excluded channels are not in the list of channels"""
        for channel in self.excluded_from_analysis:
            if channel not in self.channels:
                log.warning(
                    f"Excluded channel {channel} not found in spec for {self.name}"
                )


def make_device_spec(config: dict) -> DeviceSpec:
    """Constructs a DeviceSpec from a dict. Throws a KeyError if any fields
    are missing."""
    return DeviceSpec(name=config['name'],
                      content_type=config['content_type'],
                      channels=config['channels'],
                      sample_rate=config['sample_rate'],
                      description=config['description'],
                      excluded_from_analysis=config.get(
                          'excluded_from_analysis', []))


def load(config_path: Path = Path(DEFAULT_CONFIG), replace: bool = False) -> Dict[str, DeviceSpec]:
    """Load the list of supported hardware for data acquisition from the given
    configuration file.

    Parameters
    ----------
        config_path - path to the devices json file
        replace - optional; if true, existing devices are replaced; if false,
            values will be overwritten or appended.
    """
    global _SUPPORTED_DEVICES

    if config_path.is_file():
        with open(config_path, 'r', encoding=DEFAULT_ENCODING) as json_file:
            specs = [make_device_spec(entry) for entry in json.load(json_file)]
            if specs and replace:
                _SUPPORTED_DEVICES.clear()
            for spec in specs:
                _SUPPORTED_DEVICES[spec.name] = spec
    return _SUPPORTED_DEVICES


def preconfigured_devices() -> Dict[str, DeviceSpec]:
    """Returns the preconfigured devices keyed by name. If no devices have yet
    been configured, loads and returns the DEFAULT_CONFIG."""
    global _SUPPORTED_DEVICES
    if not _SUPPORTED_DEVICES:
        load()
    return _SUPPORTED_DEVICES


def preconfigured_device(name: str, strict: bool = True) -> DeviceSpec:
    """Retrieve the DeviceSpec with the given name. An exception is raised
    if the device is not found."""
    device = preconfigured_devices().get(name, None)
    if strict and not device:
        current = ', '.join(
            [f"'{key}'" for key, _ in preconfigured_devices().items()])
        msg = (
            f"Device not found: {name}."
            "\n\n"
            f"The current list of devices includes the following: {current}."
            "\n"
            "You may register new devices using the device module `register` function or in bulk"
            " using `load`.")
        raise ValueError(msg)
    return device


def with_content_type(content_type: str) -> List[DeviceSpec]:
    """Retrieve the list of DeviceSpecs with the given content_type."""
    return [
        spec for spec in preconfigured_devices().values()
        if spec.content_type == content_type
    ]


def register(device_spec: DeviceSpec):
    """Register the given DeviceSpec."""
    config = preconfigured_devices()
    config[device_spec.name] = device_spec



# pylint: disable=no-name-in-module,missing-docstring,too-few-public-methods
import sys
from pathlib import Path
from PyQt5.QtWidgets import QApplication, QWidget, QFileDialog, QDesktopWidget
from bcipy.preferences import preferences

DEFAULT_FILE_TYPES = "All Files (*)"


class FileDialog(QWidget):
    """GUI window that prompts the user to select a file."""

    def __init__(self):
        super().__init__()
        self.title = 'PyQt5 file dialogs - pythonspot.com'
        self.width = 640
        self.height = 480

        # Center on screen
        self.resize(self.width, self.height)
        frame_geom = self.frameGeometry()
        frame_geom.moveCenter(QDesktopWidget().availableGeometry().center())
        self.move(frame_geom.topLeft())

        # The native dialog may prevent the selection from closing after a
        # directory is selected.
        self.options = QFileDialog.Options()
        self.options |= QFileDialog.DontUseNativeDialog

    def ask_file(self, file_types: str = DEFAULT_FILE_TYPES, directory: str = "") -> str:
        """Opens a file dialog window.
        Returns
        -------
        path or None
        """
        filename, _ = QFileDialog.getOpenFileName(self,
                                                  "Select File",
                                                  directory,
                                                  file_types,
                                                  options=self.options)
        return filename

    def ask_directory(self, directory: str = "") -> str:
        """Opens a dialog window to select a directory.

        Returns
        -------
        path or None
        """
        return QFileDialog.getExistingDirectory(self,
                                                "Select Directory",
                                                directory=directory,
                                                options=self.options)


def ask_filename(file_types: str = DEFAULT_FILE_TYPES, directory: str = "") -> str:
    """Prompt for a file.

    Parameters
    ----------
    - file_types : optional file type filters; Examples: 'Text files (*.txt)'
    or 'Image files (*.jpg *.gif)' or '*.csv;;*.pkl'
    - directory : optional directory

    Returns
    -------
    path to file or None if the user cancelled the dialog.
    """
    app = QApplication(sys.argv)
    dialog = FileDialog()
    directory = directory or preferences.last_directory
    filename = dialog.ask_file(file_types, directory)

    # update last directory preference
    path = Path(filename)
    if filename and path.is_file():
        preferences.last_directory = str(path.parent)

    # Alternatively, we could use `app.closeAllWindows()`
    app.quit()

    return filename


def ask_directory() -> str:
    """Prompt for a directory.

    Returns
    -------
    path to directory or None if the user cancelled the dialog.
    """
    app = QApplication(sys.argv)

    dialog = FileDialog()
    directory = ''
    if preferences.last_directory:
        directory = str(Path(preferences.last_directory).parent)
    name = dialog.ask_directory(directory)
    if name and Path(name).is_dir():
        preferences.last_directory = name
    # Alternatively, we could use `app.closeAllWindows()`
    app.quit()

    return name



import logging
import os
from dataclasses import dataclass
from enum import Enum
from typing import Callable, Dict, List, Optional, Tuple

from psychopy import core, visual

from bcipy.config import DEFAULT_ENCODING
from bcipy.helpers.clock import Clock
from bcipy.helpers.exceptions import BciPyCoreException
from bcipy.helpers.parameters import Parameters
from bcipy.helpers.stimuli import resize_image

log = logging.getLogger(__name__)

NONE_VALUES = ['0', '0.0']

CALIBRATION_IMAGE_PATH = 'bcipy/static/images/testing/white.png'
MOCK_TRIGGER_DATA = '''starting_offset offset -3635340.451487373
P prompt 3635343.155564679
+ fixation 3635344.159978495
O nontarget 3635344.663714144
Z nontarget 3635344.917679911
B nontarget 3635345.172036632
N nontarget 3635345.425718764
K nontarget 3635345.679572615
U nontarget 3635345.932683132
J nontarget 3635346.186251469
P target 3635346.439412578
_ nontarget 3635346.692164597
T nontarget 3635346.946281828
daq_sample_offset system 293.6437802789733
'''


class CalibrationType(Enum):
    """Calibration Type.

    Enum to define the supported calibration trigger types.
    """
    TEXT = 'text'
    IMAGE = 'image'

    @classmethod
    def list(cls):
        """Returns all enum values as a list"""
        return list(map(lambda c: c.value, cls))


class TriggerCallback:
    timing = None
    first_time = True

    def callback(self, clock: Clock, stimuli: str) -> None:
        if self.first_time:
            self.timing = [stimuli, clock.getTime()]
            self.first_time = False

    def reset(self):
        self.timing = None
        self.first_time = True


def _calibration_trigger(experiment_clock: Clock,
                         trigger_type: str = CalibrationType.TEXT.value,
                         trigger_name: str = 'calibration',
                         trigger_time: float = 1,
                         display=None,
                         on_trigger=None) -> List[tuple]:
    """Calibration Trigger.

    Outputs triggers for the purpose of calibrating data and stimuli.
    This is an ongoing difficulty between OS, DAQ devices and stimuli type. This
    code aims to operationalize the approach to finding the correct DAQ samples in
    relation to our trigger code.

    PARAMETERS
    ---------
    experiment_clock(clock): clock with getTime() method, which is used in the code
        to report timing of stimuli
    trigger_type(string): type of trigger that is desired (text, image, etc)
    trigger_name(string): name of the trigger used for callbacks / labeling
    trigger_time(float): time to display the trigger. Can also be used as a buffer.
    display(DisplayWindow): a window that can display stimuli. Currently, a Psychopy window.
    on_trigger(function): optional callback; if present gets called
                when the calibration trigger is fired; accepts a single
                parameter for the timing information.
    Return:
        timing(array): timing values for the calibration triggers to be written to trigger file or
                used to calculate offsets.
    """
    trigger_callback = TriggerCallback()

    # catch invalid trigger types
    if trigger_type not in CalibrationType.list():
        msg = f'Trigger type=[{trigger_type}] not implemented'
        log.exception(msg)
        raise BciPyCoreException(msg)

    if not display:
        msg = f'Calibration type=[{trigger_type}] requires a display'
        log.exception(msg)
        raise BciPyCoreException(msg)

    if trigger_type == CalibrationType.IMAGE.value:
        calibration_box = visual.ImageStim(
            win=display,
            image=CALIBRATION_IMAGE_PATH,
            pos=(-.5, -.5),
            mask=None,
            ori=0.0)
        calibration_box.size = resize_image(CALIBRATION_IMAGE_PATH, display.size, 0.75)

        display.callOnFlip(trigger_callback.callback, experiment_clock, trigger_name)
        if on_trigger is not None:
            display.callOnFlip(on_trigger, trigger_name)

        calibration_box.draw()
        display.flip()

    elif trigger_type == CalibrationType.TEXT.value:
        calibration_text = visual.TextStim(display, text='')

        display.callOnFlip(trigger_callback.callback, experiment_clock,
                           trigger_name)
        if on_trigger is not None:
            display.callOnFlip(on_trigger, trigger_name)

        calibration_text.draw()
        display.flip()

    core.wait(trigger_time)
    return trigger_callback.timing


def trigger_durations(params: Parameters) -> Dict[str, float]:
    """Duration for each type of trigger given in seconds."""
    return {
        'offset': 0.0,
        'preview': params['preview_inquiry_length'],
        'fixation': params['time_fixation'],
        'prompt': params['time_prompt'],
        'nontarget': params['time_flash'],
        'target': params['time_flash']
    }


class TriggerType(Enum):
    """
    Enum for the primary types of Triggers.
    """

    NONTARGET = "nontarget"
    TARGET = "target"
    FIXATION = "fixation"
    PROMPT = "prompt"
    SYSTEM = "system"
    OFFSET = "offset"
    EVENT = "event"
    PREVIEW = "preview"

    @classmethod
    def list(cls) -> List[str]:
        """Returns all enum values as a list"""
        return list(map(lambda c: c.value, cls))

    def __str__(self) -> str:
        return f'{self.value}'


@dataclass(frozen=True)
class Trigger:
    """
    Object that encompasses data for a single trigger instance.
    """

    label: str
    type: TriggerType
    time: float

    def __repr__(self):
        return f'Trigger: label=[{self.label}] type=[{self.type}] time=[{self.time}]'

    def with_offset(self, offset: float):
        """Construct a copy of this Trigger with the offset adjusted."""
        return Trigger(self.label, self.type, self.time + offset)

    @classmethod
    def from_list(cls, lst: List[str]):
        """Constructs a Trigger from a serialized representation.

        Parameters
        ----------
            lst - serialized representation [label, type, stamp]
        """
        assert len(lst) == 3, "Input must have a label, type, and stamp"
        return cls(lst[0], TriggerType(lst[1]), float(lst[2]))


class FlushFrequency(Enum):
    """
    Enum that defines how often list of Triggers will be written and dumped.
    """

    EVERY = "flush after every trigger addition"
    END = "flush at end of session"


class TriggerHandler:
    """
    Class that contains methods to work with Triggers, including adding and
    writing triggers and loading triggers from a txt file.
    """

    encoding = DEFAULT_ENCODING

    def __init__(self,
                 path: str,
                 file_name: str,
                 flush: FlushFrequency):
        self.path = path
        self.file_name = f'{file_name}.txt' if not file_name.endswith('.txt') else file_name
        self.flush = flush
        self.triggers = []
        self.file_path = f'{self.path}/{self.file_name}'
        self.flush = flush
        self.triggers = []

        if os.path.exists(self.file_name):
            raise Exception(f"[{self.file_name}] already exists, any writing "
                            "will overwrite data in the existing file.")

        self.file = open(self.file_path, 'w+', encoding=self.encoding)

    def close(self) -> None:
        """Close.

        Ensures all data is written and file is closed properly.
        """
        self.write()
        self.file.close()

    def write(self) -> None:
        """
        Writes current Triggers in self.triggers[] to .txt file in self.file_name.
        File writes in the format "label, targetness, time".
        """

        for trigger in self.triggers:
            self.file.write(f'{trigger.label} {trigger.type.value} {trigger.time}\n')

        self.triggers = []

    @staticmethod
    def read_text_file(path: str) -> Tuple[List[Trigger], float]:
        """Read Triggers from the given text file.
        Parameters
        ----------
            path - trigger (.txt) file to read
        Returns
        -------
            triggers, offset
        """
        if not path.endswith('.txt') or not os.path.exists(path):
            raise FileNotFoundError(f'Valid triggers .txt file not found at [{path}].')

        with open(path, encoding=TriggerHandler.encoding) as raw_txt:
            triggers = []
            for i, line in enumerate(raw_txt):
                try:
                    trg = Trigger.from_list(line.split())
                    triggers.append(trg)
                except (AssertionError, ValueError) as trg_error:
                    raise BciPyCoreException(
                        f'Error reading trigger on line {i+1} of {path}: {trg_error}') from trg_error

        # find next offset values in list and return or create a trigger with offset of 0.0
        offset = next(
            filter(lambda trg: trg.type == TriggerType.OFFSET, triggers),
            Trigger('starting_offset', TriggerType.OFFSET, 0.0))
        triggers = [trg for trg in triggers if trg.type != TriggerType.SYSTEM]

        return triggers, offset.time

    @staticmethod
    def load(path: str,
             offset: Optional[float] = 0.0,
             exclusion: Optional[List[TriggerType]] = None) -> List[Trigger]:
        """
        Loads a list of triggers from a .txt of triggers.

        Exclusion based on type only (ex. exclusion=[TriggerType.Fixation])

        1. Checks if .txt file exists at path
        2. Loads the triggers data as a list of lists
        3. If offset provided, adds it to the time as float
        4. If exclusion provided, filters those triggers
        5. Casts all loaded and modified triggers to Trigger
        6. Returns as a List[Triggers]

        Parameters
        ----------
        path (str): name or file path of .txt trigger file to be loaded.
            Input string must include file extension (.txt).
        offset (Optional float): if desired, time offset for all loaded triggers,
            positive number for adding time, negative number for subtracting time.
        exclusion (Optional List[TriggerType]): if desired, list of TriggerType's
            to be removed from the loaded trigger list.

        Returns
        -------
            List of Triggers from loaded .txt file with desired modifications
        """

        # Checking for file with given path, with or without .txt
        triggers, system_offset = TriggerHandler.read_text_file(path)
        excluded_types = exclusion or []
        total_offset = offset + system_offset
        return [
            trg.with_offset(total_offset) for trg in triggers
            if trg.type not in excluded_types
        ]

    def add_triggers(self, triggers: List[Trigger]) -> List[Trigger]:
        """
        Adds provided list of Triggers to self.triggers.

        Parameters
        ----------
        triggers (List[Triggers]): list of Trigger objects to be added to the
            handler's list of Triggers (self.triggers).

        Returns
        -------
            Returns list of Triggers currently part of Handler
        """
        self.triggers.extend(triggers)

        if self.flush is FlushFrequency.EVERY:
            self.write()

        return self.triggers


def convert_timing_triggers(timing: List[tuple], target_stimuli: str, trigger_type: Callable) -> List[Trigger]:
    """Convert Stimuli Times to Triggers.

    Using the stimuli presentation times provided by the display, convert them into BciPy Triggers.
    """
    return [
        Trigger(symbol, trigger_type(symbol, target_stimuli, i), time) for i, (symbol, time) in enumerate(timing)
    ]


def trigger_decoder(trigger_path: str, remove_pre_fixation: bool = True, offset: float = 0.0,
                    exclusion: List[TriggerType] = []) -> Tuple[list, list, list]:
    """Trigger Decoder.

    Given a path to trigger data, this method loads valid Triggers and returns their type, timing and label.

    PARAMETERS
    ----------
    :param: trigger_path: path to triggers file
    :param: remove_pre_fixation: boolean to determine whether any stimuli before a fixation + system should be removed
    :param: offset: additional offset value to apply to triggers. If a valid 'starting_offset' present in the trigger
        file this will be applied be default.
    :param: exclusion [Optional]: any TriggerTypes to be filtered from data returned
    :return: tuple: trigger_type, trigger_timing, trigger_label
    """
    if remove_pre_fixation:
        exclusion += [TriggerType.FIXATION, TriggerType.PROMPT, TriggerType.SYSTEM, TriggerType.OFFSET]
    else:
        exclusion += [TriggerType.SYSTEM]
    triggers = TriggerHandler.load(trigger_path, offset=offset, exclusion=exclusion)

    # from the stimuli array, pull out the symbol information
    trigger_type = [trigger.type.value for trigger in triggers]
    trigger_label = [trigger.label for trigger in triggers]
    trigger_timing = [trigger.time for trigger in triggers]

    return trigger_type, trigger_timing, trigger_label



import unittest

from bcipy.helpers.symbols import alphabet


class TestAlphabet(unittest.TestCase):
    def test_alphabet_text(self):
        parameters = {}

        parameters['is_txt_stim'] = True

        alp = alphabet(parameters)

        self.assertEqual(alp, [
            'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
            'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
            '<', '_'
        ])

    def test_alphabet_images(self):
        parameters = {}
        parameters['is_txt_stim'] = False
        parameters['path_to_presentation_images'] = ('bcipy/static/images/'
                                                     'rsvp/')

        alp = alphabet(parameters)

        self.assertNotEqual(alp, [
            'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
            'N', 'O', 'P', 'R', 'S', 'T', 'U', 'V', 'Y', 'Z', '<', '_'
        ])



import logging

from abc import ABC, abstractmethod


class Task(ABC):
    """Task.

    Base class for BciPy tasks.
    """

    def __init__(self):
        super(Task, self).__init__()
        self.logger = logging.getLogger(__name__)

    def configure(self):
        ...

    @abstractmethod
    def execute(self):
        ...

    @abstractmethod
    def name(self):
        ...



import glob
import unittest

from os import path
from mockito import any, mock, unstub, verify, when

from psychopy import core
import numpy as np
import sounddevice as sd
import soundfile as sf

import collections as cnt
from bcipy.helpers.exceptions import BciPyCoreException

from bcipy.helpers.stimuli import (
    alphabetize,
    best_case_rsvp_inq_gen,
    best_selection,
    calibration_inquiry_generator,
    DEFAULT_FIXATION_PATH,
    get_fixation,
    TrialReshaper,
    InquiryReshaper,
    jittered_timing,
    play_sound,
    distributed_target_positions,
    soundfiles,
    StimuliOrder,
    ssvep_to_code,
    TargetPositions,
    update_inquiry_timing,
)

MOCK_FS = 44100


class TestStimuliGeneration(unittest.TestCase):
    """This is Test Case for Stimuli Generated via BciPy."""

    alp = [
        'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
        'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
        '<', '_'
    ]

    def tearDown(self):
        unstub()

    def test_calibration_inquiry_generator_with_jitter(self):
        stim_number = 10
        stim_length = 10
        stim_timing = [0.5, 1, 2]
        stim_jitter = 1

        max_jitter = stim_timing[-1] + stim_jitter
        min_jitter = stim_timing[-1] - stim_jitter
        inquiries, inq_timings, inq_colors = calibration_inquiry_generator(
            self.alp,
            timing=stim_timing,
            stim_number=stim_number,
            stim_length=stim_length,
            jitter=stim_jitter)

        self.assertEqual(
            len(inquiries), stim_number,
            'Should have produced the correct number of inquiries')
        self.assertEqual(len(inq_timings), stim_number)
        self.assertEqual(len(inq_colors), stim_number)

        inq_strings = []
        for inq in inquiries:
            self.assertEqual(
                len(inq), stim_length + 2,
                ('inquiry should include the correct number of choices as ',
                 'well as the target and cross.'))
            choices = inq[2:]
            self.assertEqual(stim_length, len(set(choices)),
                             'All choices should be unique')

            # create a string of the options
            inq_strings.append(''.join(choices))

        # ensure timing is jittered
        for j in inq_timings:
            inq_timing = j[2:]  # remove the target presentaion and cross
            for inq_time in inq_timing:
                self.assertTrue(min_jitter <= inq_time <= max_jitter,
                                'Timing should be jittered and within the correct range')

            self.assertTrue(
                len(set(inq_timing)) > 1, 'All choices should be unique')

        self.assertEqual(
            len(inquiries), len(set(inq_strings)),
            'All inquiries should be different')

    def test_calibration_inquiry_generator_random_order(self):
        """Test generation of random inquiries"""
        stim_number = 10
        stim_length = 10
        inquiries, inq_timings, inq_colors = calibration_inquiry_generator(
            self.alp,
            stim_number=stim_number,
            stim_length=stim_length,
            stim_order=StimuliOrder.RANDOM)

        self.assertEqual(
            len(inquiries), stim_number,
            'Should have produced the correct number of inquiries')
        self.assertEqual(len(inq_timings), stim_number)
        self.assertEqual(len(inq_colors), stim_number)

        inq_strings = []
        for inq in inquiries:
            self.assertEqual(
                len(inq), stim_length + 2,
                ('inquiry should include the correct number of choices as ',
                 'well as the target and cross.'))
            choices = inq[2:]
            self.assertEqual(stim_length, len(set(choices)),
                             'All choices should be unique')

            # create a string of the options
            inq_strings.append(''.join(choices))

        self.assertEqual(
            len(inquiries), len(set(inq_strings)),
            'All inquiries should be different')

    def test_calibration_inquiry_generator_alphabetical_order(self):
        """Test generation of random inquiries"""
        stim_number = 10
        stim_length = 10
        inquiries, inq_timings, inq_colors = calibration_inquiry_generator(
            self.alp,
            stim_number=stim_number,
            stim_length=stim_length,
            stim_order=StimuliOrder.ALPHABETICAL)

        self.assertEqual(
            len(inquiries), stim_number,
            'Should have produced the correct number of inquiries')
        self.assertEqual(len(inq_timings), stim_number)
        self.assertEqual(len(inq_colors), stim_number)

        inq_strings = []
        for inq in inquiries:
            self.assertEqual(
                len(inq), stim_length + 2,
                ('inquiry should include the correct number of choices as ',
                 'well as the target and cross.'))
            choices = inq[2:]
            self.assertEqual(stim_length, len(set(choices)),
                             'All choices should be unique')
            self.assertEqual(alphabetize(choices), choices)

            # create a string of the options
            inq_strings.append(''.join(choices))

        self.assertEqual(
            len(inquiries), len(set(inq_strings)),
            'All inquiries should be different')

    def test_calibration_inquiry_generator_distributed_targets(self):
        """Test generation of inquiries with distributed target positions"""
        stim_number = 10
        stim_length = 10
        nontarget_inquiries = 10
        inquiries, inq_timings, inq_colors = calibration_inquiry_generator(
            self.alp,
            stim_number=stim_number,
            stim_length=stim_length,
            stim_order=StimuliOrder.RANDOM,
            target_positions=TargetPositions.DISTRIBUTED,
            nontarget_inquiries=nontarget_inquiries)

        self.assertEqual(
            len(inquiries), stim_number,
            'Should have produced the correct number of inquiries')
        self.assertEqual(len(inq_timings), stim_number)
        self.assertEqual(len(inq_colors), stim_number)

        inq_strings = []
        for inq in inquiries:
            self.assertEqual(
                len(inq), stim_length + 2,
                ('inquiry should include the correct number of choices as ',
                 'well as the target and cross.'))
            choices = inq[2:]
            self.assertEqual(stim_length, len(set(choices)),
                             'All choices should be unique')

            # create a string of the options
            inq_strings.append(''.join(choices))

        self.assertEqual(
            len(inquiries), len(set(inq_strings)),
            'All inquiries should be different')

    def test_calibration_inquiry_generator_distributed_targets_alphabetical(self):
        """Test generation of inquiries with distributed target positions"""
        stim_number = 10
        stim_length = 10
        nontarget_inquiries = 20
        inquiries, inq_timings, inq_colors = calibration_inquiry_generator(
            self.alp,
            stim_number=stim_number,
            stim_length=stim_length,
            stim_order=StimuliOrder.ALPHABETICAL,
            target_positions=TargetPositions.DISTRIBUTED,
            nontarget_inquiries=nontarget_inquiries)

        self.assertEqual(
            len(inquiries), stim_number,
            'Should have produced the correct number of inquiries')
        self.assertEqual(len(inq_timings), stim_number)
        self.assertEqual(len(inq_colors), stim_number)

        inq_strings = []
        for inq in inquiries:
            self.assertEqual(
                len(inq), stim_length + 2,
                ('inquiry should include the correct number of choices as ',
                 'well as the target and cross.'))
            choices = inq[2:]
            self.assertEqual(stim_length, len(set(choices)),
                             'All choices should be unique')

            # create a string of the options
            inq_strings.append(''.join(choices))

        self.assertEqual(
            len(inquiries), len(set(inq_strings)),
            'All inquiries should be different')

    def test_calibration_inquiry_generator_distributed_targets_no_nontargets(self):
        """Test generation of inquiries with distributed target positions and no nontarget inquiries."""
        stim_number = 10
        stim_length = 10
        nontarget_inquiries = 0
        inquiries, inq_timings, inq_colors = calibration_inquiry_generator(
            self.alp,
            stim_number=stim_number,
            stim_length=stim_length,
            stim_order=StimuliOrder.RANDOM,
            target_positions=TargetPositions.DISTRIBUTED,
            nontarget_inquiries=nontarget_inquiries)

        self.assertEqual(
            len(inquiries), stim_number,
            'Should have produced the correct number of inquiries')
        self.assertEqual(len(inq_timings), stim_number)
        self.assertEqual(len(inq_colors), stim_number)

        inq_strings = []
        for inq in inquiries:
            self.assertEqual(
                len(inq), stim_length + 2,
                ('inquiry should include the correct number of choices as ',
                 'well as the target and cross.'))
            choices = inq[2:]
            self.assertEqual(stim_length, len(set(choices)),
                             'All choices should be unique')

            # create a string of the options
            inq_strings.append(''.join(choices))

        self.assertEqual(
            len(inquiries), len(set(inq_strings)),
            'All inquiries should be different')

    def test_calibration_inquiry_generator_distributed_targets_positions(self):
        """Test generation of distributed target positions with nontarget inquiries."""

        stim_number = 11
        stim_length = 10
        nontarget_inquiries = 10

        nontarget_inquiry = (int)(stim_number * (nontarget_inquiries / 100))
        target_inquiries = stim_number - nontarget_inquiry
        num_target_inquiries = (int)(target_inquiries / stim_length)

        targets = distributed_target_positions(stim_number=stim_number,
                                               stim_length=stim_length,
                                               nontarget_inquiries=nontarget_inquiries)

        self.assertEqual(
            len(targets), stim_number,
            'Should have produced the correct number of targets for inquiries.')

        # count how many times each target position is used
        count = cnt.Counter()
        for pos in targets:
            count[pos] += 1

        # make sure position counts are equally distributed, including non-target
        for i in count:
            self.assertTrue(num_target_inquiries <= count[i] <= num_target_inquiries + 1)

    def test_calibration_inquiry_generator_distributed_targets_positions_half_nontarget(self):
        """Test generation of distributed target positions with half being nontarget inquiries."""

        stim_number = 120
        stim_length = 9
        nontarget_inquiries = 50

        nontarget_inquiry = (int)(stim_number * (nontarget_inquiries / 100))
        target_inquiries = stim_number - nontarget_inquiry
        num_target_inquiries = (int)(target_inquiries / stim_length)

        targets = distributed_target_positions(stim_number=stim_number,
                                               stim_length=stim_length,
                                               nontarget_inquiries=nontarget_inquiries)

        self.assertEqual(
            len(targets), stim_number,
            'Should have produced the correct number of targets for inquiries.')

        # count how many times each target position is used
        count = cnt.Counter()
        for pos in targets:
            count[pos] += 1

        # make sure target position counts are equally distributed
        for i in count:
            if i is not None:
                self.assertTrue(num_target_inquiries <= count[i] <= num_target_inquiries + 1)

        # make sure correct number of non-target inquiries
        self.assertEqual(count[None], nontarget_inquiry,
                         'Should have produced 50 percent of 120 non-target positions.')

    def test_calibration_inquiry_generator_distributed_targets_positions_no_nontargets(self):
        """Test generation of distributed target positions with no nontarget inquiries."""

        stim_number = 50
        stim_length = 11
        nontarget_inquiries = 0

        nontarget_inquiry = (int)(stim_number * (nontarget_inquiries / 100))
        target_inquiries = stim_number - nontarget_inquiry
        num_target_inquiries = (int)(target_inquiries / stim_length)

        targets = distributed_target_positions(stim_number=stim_number,
                                               stim_length=stim_length,
                                               nontarget_inquiries=nontarget_inquiries)

        self.assertEqual(
            len(targets), stim_number,
            'Should have produced the correct number of targets for inquiries.')

        # count how many times each target position is used
        count = cnt.Counter()
        for pos in targets:
            count[pos] += 1

        # make sure target position counts are equally distributed
        for i in count:
            if i is not None:
                self.assertTrue(num_target_inquiries <= count[i] <= num_target_inquiries + 1)

        # make sure there are no non-target inquiries
        self.assertEqual(count[None], 0,
                         'Should have produced no non-target positions.')

    def test_calibration_inquiry_generator_distributed_targets_all_nontargets(self):
        """Test generation of distributed target positions with all inquiries being non-target."""

        stim_number = 100
        stim_length = 6
        nontarget_inquiries = 100

        nontarget_inquiry = (int)(stim_number * (nontarget_inquiries / 100))
        target_inquiries = stim_number - nontarget_inquiry
        num_target_inquiries = (int)(target_inquiries / stim_length)

        targets = distributed_target_positions(
            stim_number=stim_number,
            stim_length=stim_length,
            nontarget_inquiries=nontarget_inquiries)

        self.assertEqual(
            len(targets), stim_number,
            'Should have produced the correct number of targets for inquiries.')

        # count how many times each target position is used
        count = cnt.Counter()
        for pos in targets:
            count[pos] += 1

        # make sure target position counts are equally distributed
        for i in count:
            if i is not None:
                self.assertTrue(num_target_inquiries <= count[i] <= num_target_inquiries + 1)

        # make sure all inquries are non-target inquiries
        self.assertEqual(count[None], stim_number,
                         'Should have produced all non-target positions.')

    def test_best_selection(self):
        """Test best_selection"""

        self.assertEqual(['a', 'c', 'e'],
                         best_selection(
                             selection_elements=['a', 'b', 'c', 'd', 'e'],
                             val=[0.3, 0.1, 0.3, 0.1, 0.2],
                             len_query=3))

        # Test equal probabilities
        self.assertEqual(['a', 'b', 'c'],
                         best_selection(
                             selection_elements=['a', 'b', 'c', 'd', 'e'],
                             val=[0.2, 0.2, 0.2, 0.2, 0.2],
                             len_query=3))

        # Test always included
        self.assertEqual(
            ['a', 'c', 'd'],
            best_selection(
                selection_elements=['a', 'b', 'c', 'd', 'e'],
                val=[0.3, 0.1, 0.3, 0.1, 0.2],
                len_query=3,
                always_included=['d']),
            'Included item should bump out the best item with the lowest val.')

        self.assertEqual(
            ['a', 'b', 'c'],
            best_selection(
                selection_elements=['a', 'b', 'c', 'd', 'e'],
                val=[0.5, 0.4, 0.1, 0.0, 0.0],
                len_query=3,
                always_included=['b']),
            'Included item should retain its position if already present')

        self.assertEqual(['a', 'b', 'e'],
                         best_selection(
                             selection_elements=['a', 'b', 'c', 'd', 'e'],
                             val=[0.5, 0.0, 0.1, 0.3, 0.0],
                             len_query=3,
                             always_included=['b', 'e']),
                         'multiple included items should be supported.')

        self.assertEqual(['d'],
                         best_selection(
                             selection_elements=['a', 'b', 'c', 'd', 'e'],
                             val=[0.5, 0.4, 0.1, 0.0, 0.0],
                             len_query=1,
                             always_included=['d', 'e']),
                         'len_query should be respected.')

        self.assertEqual(['a', 'b', 'c'],
                         best_selection(
            selection_elements=['a', 'b', 'c', 'd', 'e'],
            val=[0.5, 0.4, 0.1, 0.0, 0.0],
            len_query=3,
            always_included=['<']),
            'should ignore items not in the set.')

    def test_best_case_inquiry_gen(self):
        """Test best_case_rsvp_inq_gen"""
        alp = ['a', 'b', 'c', 'd', 'e', 'f', 'g']
        n = 5
        samples, times, colors = best_case_rsvp_inq_gen(
            alp=alp,
            session_stimuli=[0.1, 0.1, 0.1, 0.2, 0.2, 0.1, 0.2],
            timing=[1, 0.2],
            color=['red', 'white'],
            stim_number=1,
            stim_length=n,
            is_txt=True)

        first_inq = samples[0]
        self.assertEqual(1, len(samples))
        self.assertEqual(n + 1, len(first_inq),
                         'Should include fixation cross.')
        self.assertEqual(len(samples), len(times))
        self.assertEqual(len(samples), len(colors))

        expected = ['+', 'a', 'b', 'd', 'e', 'g']
        for letter in expected:
            self.assertTrue(letter in first_inq)

        self.assertNotEqual(expected, first_inq, 'Should be in random order.')
        self.assertEqual([1] + ([0.2] * n), times[0])
        self.assertEqual(['red'] + (['white'] * n), colors[0])

    def test_best_case_inquiry_gen_with_inq_constants(self):
        """Test best_case_rsvp_inq_gen with inquiry constants"""

        alp = ['a', 'b', 'c', 'd', 'e', 'f', 'g']
        n = 5

        with self.assertRaises(
                Exception, msg='Constants should be in the alphabet'):
            best_case_rsvp_inq_gen(
                alp=alp,
                session_stimuli=[0.1, 0.1, 0.1, 0.2, 0.2, 0.1, 0.2],
                inq_constants=['<'])

        alp = ['a', 'b', 'c', 'd', 'e', 'f', 'g', '<']
        samples, times, colors = best_case_rsvp_inq_gen(
            alp=alp,
            session_stimuli=[0.1, 0.1, 0.1, 0.2, 0.2, 0.1, 0.2, 0.0],
            stim_number=1,
            stim_length=n,
            is_txt=True,
            inq_constants=['<'])

        first_inq = samples[0]
        self.assertEqual(1, len(samples))
        self.assertEqual(n + 1, len(first_inq),
                         'Should include fixation cross.')
        self.assertEqual(len(samples), len(times))
        self.assertEqual(len(samples), len(colors))

        expected = ['+', 'a', 'd', 'e', 'g', '<']
        for letter in expected:
            self.assertTrue(letter in first_inq)

        self.assertNotEqual(expected, first_inq, 'Should be in random order.')
        self.assertEqual([1] + ([0.2] * n), times[0])
        self.assertEqual(['red'] + (['white'] * n), colors[0])


class TestJitteredTiming(unittest.TestCase):

    def test_jittered_timing_returns_correct_number_of_stims(self):
        time = 1
        jitter = 0.24
        stim_number = 10

        resp = jittered_timing(time, jitter, stim_number)

        self.assertEqual(stim_number, len(resp))
        self.assertIsInstance(resp, list)

    def test_jittered_timing_with_jitter_within_defined_limits(self):
        time = 1
        jitter = 0.25
        stim_number = 100

        max_jitter = time + jitter
        min_jitter = time - jitter

        resp = jittered_timing(time, jitter, stim_number)

        for r in resp:
            self.assertTrue(min_jitter <= r <= max_jitter)

    def test_jittered_timing_throw_exception_when_jitter_greater_than_time(self):
        # to prevent 0 values we prevent the jitter from being greater than the time
        time = 1
        jitter = 1.5
        stim_number = 100

        with self.assertRaises(Exception, msg='Jitter should be less than stimuli time'):
            jittered_timing(time, jitter, stim_number)


class TestGetFixation(unittest.TestCase):

    def test_text_fixation(self):
        expected = '+'
        response = get_fixation(is_txt=True)
        self.assertEqual(expected, response)

    def test_image_fixation_uses_default(self):
        expected = DEFAULT_FIXATION_PATH
        response = get_fixation(is_txt=False)
        self.assertEqual(expected, response)


class TestAlphabetize(unittest.TestCase):

    def setUp(self) -> None:
        self.list_to_alphabetize = ['Z', 'Q', 'A', 'G']

    def test_alphabetize(self):
        expected = ['A', 'G', 'Q', 'Z']
        response = alphabetize(self.list_to_alphabetize)
        self.assertEqual(expected, response)

    def test_alphabetize_image_name(self):
        list_of_images = ['testing.png', '_ddtt.jpeg', 'bci_image.png']
        expected = ['bci_image.png', 'testing.png', '_ddtt.jpeg']
        response = alphabetize(list_of_images)
        self.assertEqual(expected, response)

    def test_alphabetize_special_characters_at_end(self):
        character = '<'
        expected = ['A', 'G', 'Q', 'Z', character]
        self.list_to_alphabetize.insert(1, character)
        response = alphabetize(self.list_to_alphabetize)
        self.assertEqual(expected, response)


class TestTrialReshaper(unittest.TestCase):
    def setUp(self):
        self.target_info = ['target', 'nontarget', 'nontarget']
        self.timing_info = [1.001, 1.2001, 1.4001]
        # make some fake eeg data
        self.channel_number = 21
        tmp_inp = np.array([range(4000)] * self.channel_number)
        # Add some channel info
        tmp_inp[:, 0] = np.transpose(np.arange(1, self.channel_number + 1, 1))
        self.eeg = tmp_inp
        self.channel_map = [1] * self.channel_number

    def test_trial_reshaper(self):
        sample_rate = 256
        trial_length_s = 0.5
        reshaped_trials, labels = TrialReshaper()(
            trial_targetness_label=self.target_info,
            timing_info=self.timing_info,
            eeg_data=self.eeg,
            sample_rate=sample_rate,
            channel_map=self.channel_map,
            poststimulus_length=trial_length_s)
        trial_length_samples = int(sample_rate * trial_length_s)
        expected_shape = (self.channel_number, len(self.target_info), trial_length_samples)
        self.assertTrue(np.all(labels == [1, 0, 0]))
        self.assertTrue(reshaped_trials.shape == expected_shape)


class TestInquiryReshaper(unittest.TestCase):

    def setUp(self):
        self.n_channel = 7
        self.trial_length = 0.5
        self.trials_per_inquiry = 3
        self.n_inquiry = 4
        self.sample_rate = 10
        self.target_info = [
            "target", "nontarget", "nontarget",
            "nontarget", "nontarget", "nontarget",
            "nontarget", "target", "nontarget",
            "nontarget", "nontarget", "target",
        ]
        self.true_labels = np.array([
            [1, 0, 0],
            [0, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ])
        self.timing_info = [
            1.4, 2.4, 3.4,
            28.1, 30.1, 38.1,
            50.1, 51.1, 52.1,
            80.1, 81.1, 82.1,
        ]
        # Note this value must be greater than the difference between first and last timing info per inquiry
        # In this case there are 3 trials per inquiry and the difference between the first and last timing info is 10s
        self.inquiry_duration_s = 10 + self.trial_length
        self.samples_per_inquiry = int(self.sample_rate * self.inquiry_duration_s)
        self.samples_per_trial = int(self.sample_rate * self.trial_length)
        # We create a wealth of samples, but only use the first 4 inquiries worth
        self.eeg = np.random.randn(self.n_channel, 10000)
        self.channel_map = [1] * self.n_channel

    def test_inquiry_reshaper(self):
        reshaped_data, labels, _ = InquiryReshaper()(
            trial_targetness_label=self.target_info,
            timing_info=self.timing_info,
            eeg_data=self.eeg,
            sample_rate=self.sample_rate,
            trials_per_inquiry=self.trials_per_inquiry,
            channel_map=self.channel_map,
            poststimulus_length=self.trial_length,
        )
        expected_shape = (self.n_channel, self.n_inquiry, self.samples_per_inquiry)
        self.assertTrue(reshaped_data.shape == expected_shape)
        self.assertTrue(np.all(labels == self.true_labels))

    def test_inquiry_reshaper_trial_extraction(self):
        timing = [[1, 3, 4], [1, 4, 5], [1, 2, 3], [4, 5, 6]]
        # make a fake eeg data array (n_channels, n_inquiry, n_samples)
        inquiries = np.random.randn(self.n_channel, self.n_inquiry, self.samples_per_inquiry)

        response = InquiryReshaper().extract_trials(
            inquiries=inquiries,
            samples_per_trial=self.samples_per_trial,
            inquiry_timing=timing,
        )
        expected_shape = (self.n_channel, (self.trials_per_inquiry * self.n_inquiry), self.samples_per_trial)
        self.assertTrue(response.shape == expected_shape)

    def test_inquiry_reshaper_trial_extraction_with_prestimulus(self):
        timing = [[2, 7, 10], [2, 7, 15], [2, 12, 15], [4, 5, 6]]
        # make a fake eeg data array (n_channels, n_inquiry, n_samples)
        inquiries = np.random.randn(self.n_channel, self.n_inquiry, self.samples_per_inquiry)
        prestimulus_samples = 1
        response = InquiryReshaper().extract_trials(
            inquiries=inquiries,
            samples_per_trial=self.samples_per_trial,
            inquiry_timing=timing,
            prestimulus_samples=prestimulus_samples,
        )
        expected_shape = (
            self.n_channel,
            (self.trials_per_inquiry * self.n_inquiry),
            (self.samples_per_trial + prestimulus_samples))
        self.assertTrue(response.shape == expected_shape)


class TestUpdateInquiryTiming(unittest.TestCase):

    def test_update_inquiry_timing(self):
        initial_timing = [[100, 200]]
        downsample = 2
        new_timing = update_inquiry_timing(
            timing=initial_timing,
            downsample=downsample,
        )
        expected_timing = [[50, 100]]
        self.assertTrue(new_timing == expected_timing)

    def test_update_inquiry_timing_with_non_int_timing_after_correction(self):
        initial_timing = [[100, 201]]
        downsample = 2
        new_timing = update_inquiry_timing(
            timing=initial_timing,
            downsample=downsample,
        )
        expected_timing = [[50, 201 // downsample]]
        self.assertTrue(new_timing == expected_timing)


class SSVEPStimuli(unittest.TestCase):

    def test_default_flicker_and_refresh_rate_return_codes(self):
        response = ssvep_to_code()
        self.assertIsInstance(response, list)
        self.assertEqual(response[0], 0)

    def test_ssvep_to_codes_returns_the_length_of_refresh_rate(self):
        refresh_rate = 40
        flicker_rate = 2
        response = ssvep_to_code(flicker_rate=flicker_rate, refresh_rate=refresh_rate)
        self.assertTrue(len(response) == refresh_rate)
        self.assertEqual(response[0], 0)
        self.assertEqual(response[-1], 1)

    def test_ssvep_to_code_raises_exception_when_refresh_rate_less_than_flicker_rate(self):
        flicker_rate = 300
        refresh_rate = 1

        with self.assertRaises(BciPyCoreException):
            ssvep_to_code(refresh_rate, flicker_rate)

    def test_when_division_of_refresh_rate_by_flicker_rate_raise_exception_if_noninteger(self):
        flicker_rate = 11
        refresh_rate = 60

        with self.assertRaises(BciPyCoreException):
            ssvep_to_code(refresh_rate, flicker_rate)

    def test_ssvep_to_code_returns_expected_codes(self):
        flicker_rate = 2
        refresh_rate = 4
        response = ssvep_to_code(flicker_rate=flicker_rate, refresh_rate=refresh_rate)
        expected_output = [0, 0, 1, 1]
        self.assertEqual(response, expected_output)

    def test_ssvep_to_code_raises_exception_when_flicker_rate_one_or_less(self):
        flicker_rate = 1
        refresh_rate = 2
        with self.assertRaises(BciPyCoreException):
            ssvep_to_code(refresh_rate, flicker_rate)

        flicker_rate = 0
        refresh_rate = 2
        with self.assertRaises(BciPyCoreException):
            ssvep_to_code(refresh_rate, flicker_rate)


class TestSoundStimuli(unittest.TestCase):

    def tearDown(self):
        unstub()

    def test_play_sound_returns_timing(self):
        # fake sound file path
        sound_file_path = 'test_sound_file_path'

        # mock the other library interactions
        when(sf).read(
            sound_file_path, dtype='float32').thenReturn(('data', MOCK_FS))
        when(sd).play(any(), any()).thenReturn(None)
        when(core).wait(any()).thenReturn(None)

        # play our test sound file
        timing = play_sound(sound_file_path)

        # assert the response is as expected
        self.assertIsInstance(timing, list)

        # verify all the expected calls happended and the expected number of times
        verify(sf, times=1).read(sound_file_path, dtype='float32')
        verify(sd, times=1).play('data', MOCK_FS)
        verify(core, times=2).wait(any())

    def test_play_sound_raises_exception_if_soundfile_cannot_read_file(self):
        # fake sound file path
        sound_file_path = 'test_sound_file_path'

        # mock the other library interactions
        when(sf).read(
            sound_file_path, dtype='float32').thenRaise(Exception(''))

        # assert it raises the exception
        with self.assertRaises(Exception):
            play_sound(sound_file_path)

        # verify all the expected calls happended and the expected number of times
        verify(sf, times=1).read(sound_file_path, dtype='float32')

    def test_play_sound_sound_callback_evokes_with_timing(self):
        # fake sound file path
        sound_file_path = 'test_sound_file_path'
        test_trigger_name = 'test_trigger_name'
        test_trigger_time = 111
        self.test_timing = [test_trigger_name, test_trigger_time]

        experiment_clock = mock()

        def mock_callback_function(clock, stimuli):
            self.assertEqual(stimuli, self.test_timing[0])

        # mock the other library interactions
        when(sf).read(
            sound_file_path, dtype='float32').thenReturn(('data', MOCK_FS))
        when(sd).play(any(), any()).thenReturn(None)
        when(core).wait(any()).thenReturn(None)
        when(experiment_clock).getTime().thenReturn(test_trigger_time)

        play_sound(
            sound_file_path,
            track_timing=True,
            sound_callback=mock_callback_function,
            trigger_name=test_trigger_name,
            experiment_clock=experiment_clock,
        )

        # verify all the expected calls happended and the expected number of times
        verify(sf, times=1).read(sound_file_path, dtype='float32')
        verify(sd, times=1).play('data', MOCK_FS)
        verify(core, times=2).wait(any())

    def test_soundfiles_generator(self):
        """Test that soundfiles function returns an cyclic generator."""

        directory = path.join('.', 'sounds')
        soundfile_paths = [
            path.join(directory, '0.wav'),
            path.join(directory, '1.wav'),
            path.join(directory, '2.wav')
        ]
        when(glob).glob(
            path.join(
                directory,
                '*.wav')).thenReturn(soundfile_paths)
        when(path).isdir(directory).thenReturn(True)

        gen = soundfiles(directory)
        self.assertEqual(next(gen), soundfile_paths[0])
        self.assertEqual(next(gen), soundfile_paths[1])
        self.assertEqual(next(gen), soundfile_paths[2])
        self.assertEqual(next(gen), soundfile_paths[0])
        self.assertEqual(next(gen), soundfile_paths[1])
        for _ in range(10):
            self.assertTrue(next(gen) in soundfile_paths)

    def test_soundfiles_generator_path_arg(self):
        """Test that soundfiles function constructs the correct path."""
        directory = path.join('.', 'sounds')
        soundfile_paths = [
            path.join(directory, '0.wav'),
            path.join(directory, '1.wav'),
            path.join(directory, '2.wav')
        ]
        when(glob).glob(
            path.join(
                directory,
                '*.wav')).thenReturn(soundfile_paths)
        when(path).isdir(directory).thenReturn(True)
        gen = soundfiles(directory)
        self.assertEqual(next(gen), soundfile_paths[0])


if __name__ == '__main__':
    unittest.main()



import logging
from pathlib import Path
from typing import Tuple

from bcipy.config import (DEFAULT_PARAMETERS_PATH, TRIGGER_FILENAME,
                          RAW_DATA_FILENAME, STATIC_AUDIO_PATH,
                          DEFAULT_DEVICE_SPEC_FILENAME)
from bcipy.preferences import preferences
from bcipy.helpers.acquisition import analysis_channels
from bcipy.helpers.load import (
    load_experimental_data,
    load_json_parameters,
    load_raw_data,
)
from bcipy.helpers.stimuli import play_sound, update_inquiry_timing
from bcipy.helpers.system_utils import report_execution_time
from bcipy.helpers.triggers import TriggerType, trigger_decoder
from bcipy.helpers.visualization import visualize_erp
from bcipy.signal.model.base_model import SignalModel
from bcipy.signal.model.pca_rda_kde import PcaRdaKdeModel
from bcipy.signal.process import filter_inquiries, get_default_transform

import numpy as np
from matplotlib.figure import Figure
from sklearn.metrics import balanced_accuracy_score
from sklearn.model_selection import train_test_split
import bcipy.acquisition.devices as devices

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format="[%(threadName)-9s][%(asctime)s][%(name)s][%(levelname)s]: %(message)s")


def subset_data(data: np.ndarray, labels: np.ndarray, test_size: float, random_state=0):
    """Performs a train/test split on the provided data and labels, accounting for
    the current shape convention (channel dimension in front, instead of batch dimension in front).

    Args:
        data (np.ndarray): Shape (channels, items, time)
        labels (np.ndarray): Shape (items,)
        test_size (float): fraction of data to be used for testing
        random_state (int, optional): fixed random seed

    Returns:
        train_data (np.ndarray): Shape (channels, train_items, time)
        test_data (np.ndarray): Shape (channels, test_items, time)
        train_labels (np.ndarray): Shape (train_items,)
        test_labels (np.ndarray): Shape (test_items,)
    """
    data = data.swapaxes(0, 1)
    train_data, test_data, train_labels, test_labels = train_test_split(
        data, labels, test_size=test_size, random_state=random_state
    )
    train_data = train_data.swapaxes(0, 1)
    test_data = test_data.swapaxes(0, 1)
    return train_data, test_data, train_labels, test_labels


@report_execution_time
def offline_analysis(
    data_folder: str = None,
    parameters: dict = {},
    alert_finished: bool = True,
    estimate_balanced_acc: bool = False,
    show_figures: bool = False,
    save_figures: bool = False,
) -> Tuple[SignalModel, Figure]:
    """Gets calibration data and trains the model in an offline fashion.
    pickle dumps the model into a .pkl folder
    Args:
        data_folder(str): folder of the data
            save all information and load all from this folder
        parameter(dict): parameters for running offline analysis
        alert_finished(bool): whether or not to alert the user offline analysis complete
        estimate_balanced_acc(bool): if true, uses another model copy on an 80/20 split to
            estimate balanced accuracy
        show_figures(bool): if true, shows ERP figures after training
        save_figures(bool): if true, saves ERP figures after training to the data folder

    How it Works:
    - reads data and information from a .csv calibration file
    - reads trigger information from a .txt trigger file
    - filters data
    - reshapes and labels the data for the training procedure
    - fits the model to the data
        - uses cross validation to select parameters
        - based on the parameters, trains system using all the data
    - pickle dumps model into .pkl file
    - generates and [optional] saves/shows the ERP figure
    - [optional] alert the user finished processing
    """

    if not data_folder:
        data_folder = load_experimental_data()

    # extract relevant session information from parameters file
    poststim_length = parameters.get("trial_length")
    prestim_length = parameters.get("prestim_length")
    trials_per_inquiry = parameters.get("stim_length")
    # The task buffer length defines the min time between two inquiries
    # We use half of that time here to buffer during transforms
    buffer = int(parameters.get("task_buffer_length") / 2)
    raw_data_file = f"{RAW_DATA_FILENAME}.csv"

    # get signal filtering information
    downsample_rate = parameters.get("down_sampling_rate")
    notch_filter = parameters.get("notch_filter_frequency")
    filter_high = parameters.get("filter_high")
    filter_low = parameters.get("filter_low")
    filter_order = parameters.get("filter_order")
    static_offset = parameters.get("static_trigger_offset")

    log.info(
        f"\nData processing settings: \n"
        f"Filter: [{filter_low}-{filter_high}], Order: {filter_order},"
        f" Notch: {notch_filter}, Downsample: {downsample_rate} \n"
        f"Poststimulus: {poststim_length}s, Prestimulus: {prestim_length}s, Buffer: {buffer}s \n"
        f"Static offset: {static_offset}"
    )

    # Load raw data
    raw_data = load_raw_data(Path(data_folder, raw_data_file))
    channels = raw_data.channels
    type_amp = raw_data.daq_type
    sample_rate = raw_data.sample_rate

    devices.load(Path(data_folder, DEFAULT_DEVICE_SPEC_FILENAME))
    device_spec = devices.preconfigured_device(raw_data.daq_type)

    # setup filtering
    default_transform = get_default_transform(
        sample_rate_hz=sample_rate,
        notch_freq_hz=notch_filter,
        bandpass_low=filter_low,
        bandpass_high=filter_high,
        bandpass_order=filter_order,
        downsample_factor=downsample_rate,
    )

    log.info(f"Channels read from csv: {channels}")
    log.info(f"Device type: {type_amp}, fs={sample_rate}")

    k_folds = parameters.get("k_folds")
    model = PcaRdaKdeModel(k_folds=k_folds)

    # Process triggers.txt files
    trigger_targetness, trigger_timing, trigger_symbols = trigger_decoder(
        offset=static_offset,
        trigger_path=f"{data_folder}/{TRIGGER_FILENAME}",
        exclusion=[TriggerType.PREVIEW, TriggerType.EVENT, TriggerType.FIXATION],
    )
    # Channel map can be checked from raw_data.csv file or the devices.json located in the acquisition module
    # The timestamp column [0] is already excluded.
    channel_map = analysis_channels(channels, device_spec)
    channels_used = [channels[i] for i, keep in enumerate(channel_map) if keep == 1]
    log.info(f'Channels used in analysis: {channels_used}')

    data, fs = raw_data.by_channel()

    inquiries, inquiry_labels, inquiry_timing = model.reshaper(
        trial_targetness_label=trigger_targetness,
        timing_info=trigger_timing,
        eeg_data=data,
        sample_rate=sample_rate,
        trials_per_inquiry=trials_per_inquiry,
        channel_map=channel_map,
        poststimulus_length=poststim_length,
        prestimulus_length=prestim_length,
        transformation_buffer=buffer,
    )

    inquiries, fs = filter_inquiries(inquiries, default_transform, sample_rate)
    inquiry_timing = update_inquiry_timing(inquiry_timing, downsample_rate)
    trial_duration_samples = int(poststim_length * fs)
    data = model.reshaper.extract_trials(inquiries, trial_duration_samples, inquiry_timing)

    # define the training classes using integers, where 0=nontargets/1=targets
    labels = inquiry_labels.flatten()

    # train and save the model as a pkl file
    log.info("Training model. This will take some time...")
    model = PcaRdaKdeModel(k_folds=k_folds)
    model.fit(data, labels)
    log.info(f"Training complete [AUC={model.auc:0.4f}]. Saving data...")

    model.save(data_folder + f"/model_{model.auc:0.4f}.pkl")
    preferences.signal_model_directory = data_folder

    # Using an 80/20 split, report on balanced accuracy
    if estimate_balanced_acc:
        train_data, test_data, train_labels, test_labels = subset_data(data, labels, test_size=0.2)
        dummy_model = PcaRdaKdeModel(k_folds=k_folds)
        dummy_model.fit(train_data, train_labels)
        probs = dummy_model.predict_proba(test_data)
        preds = probs.argmax(-1)
        score = balanced_accuracy_score(test_labels, preds)
        log.info(f"Balanced acc with 80/20 split: {score}")
        del dummy_model, train_data, test_data, train_labels, test_labels, probs, preds

    figure_handles = visualize_erp(
        raw_data,
        channel_map,
        trigger_timing,
        labels,
        poststim_length,
        transform=default_transform,
        plot_average=True,
        plot_topomaps=True,
        save_path=data_folder if save_figures else None,
        show=show_figures
    )
    if alert_finished:
        play_sound(f"{STATIC_AUDIO_PATH}/{parameters['alert_sound_file']}")
    return model, figure_handles


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--data_folder", default=None)
    parser.add_argument("-p", "--parameters_file", default=DEFAULT_PARAMETERS_PATH)
    parser.add_argument("-s", "--save_figures", action="store_true")
    parser.add_argument("-v", "--show_figures", action="store_true")
    parser.add_argument("--alert", dest="alert", action="store_true")
    parser.add_argument("--balanced-acc", dest="balanced", action="store_true")
    parser.set_defaults(alert=False)
    parser.set_defaults(balanced=False)
    parser.set_defaults(save_figures=False)
    parser.set_defaults(show_figures=False)
    args = parser.parse_args()

    log.info(f"Loading params from {args.parameters_file}")
    parameters = load_json_parameters(args.parameters_file, value_cast=True)

    offline_analysis(
        args.data_folder,
        parameters,
        alert_finished=args.alert,
        estimate_balanced_acc=args.balanced,
        save_figures=args.save_figures,
        show_figures=args.show_figures)
    log.info("Offline Analysis complete.")



"""Helper functions for working with the acquisition module"""
import logging
import subprocess
import time
from typing import Dict, List, Tuple, Optional

import numpy as np

from bcipy.acquisition.devices import (DeviceSpec, preconfigured_device,
                                       with_content_type)
from bcipy.acquisition import (LslAcquisitionClient, await_start,
                               LslDataServer, discover_device_spec,
                               ClientManager)
from bcipy.helpers.save import save_device_specs
from bcipy.config import BCIPY_ROOT, RAW_DATA_FILENAME, DEFAULT_DEVICE_SPEC_FILENAME as spec_name

log = logging.getLogger(__name__)


def init_eeg_acquisition(
        parameters: dict,
        save_folder: str,
        server: bool = False) -> Tuple[ClientManager, List[LslDataServer]]:
    """Initialize EEG Acquisition.

    Initializes a client that connects with the EEG data source and begins
    data collection.

    Parameters
    ----------
        parameters : dict
            configuration details regarding the device type and other relevant
            connection information.
             {
               "acq_device": str
             }
        save_folder : str
            path to the folder where data should be saved.
        server : bool, optional
            optionally start a mock data server that streams random data.
            For discovered devices (ex. acquisition_mode: 'EEG'), the first
            matching preconfigured device will be mocked.

    Returns
    -------
        (client_manager, servers) tuple
    """

    servers = []
    manager = ClientManager()

    for stream_type in stream_types(parameters['acq_mode']):
        content_type, device_name = parse_stream_type(stream_type)

        if server:
            server_device_spec = server_spec(content_type, device_name)
            log.info(
                f"Generating mock device data for {server_device_spec.name}")
            dataserver = LslDataServer(server_device_spec)
            servers.append(dataserver)
            # Start the server before init_device so it is discoverable.
            await_start(dataserver)

        device_spec = init_device(content_type, device_name)
        raw_data_name = f'{RAW_DATA_FILENAME}.csv' if content_type == 'EEG' else None

        client = init_lsl_client(parameters, device_spec, save_folder,
                                 raw_data_name)
        manager.add_client(client)

    manager.start_acquisition()

    if parameters['acq_show_viewer']:
        viewer_screen = 1 if int(parameters['stim_screen']) == 0 else 0
        start_viewer(display_screen=viewer_screen,
                     parameter_location=parameters['parameter_location'])

    save_device_specs(manager.device_specs, save_folder, spec_name)

    return (manager, servers)


def init_device(content_type: str,
                device_name: Optional[str] = None) -> DeviceSpec:
    """Initialize a DeviceSpec for the given content type.

    If a device_name is provided, the DeviceSpec will be looked up from the list
    of preconfigured devices, otherwise, BciPy will attempt to discover a device
    with the given type streaming on the network.

    If a discovered device is found in devices.json, the provided configuration
    will override any discovered fields.

    Parameters
    ----------
        content_type - LSL content type (EEG, Gaze, etc).
        device_name - optional; name of the device. If provided, the DeviceSpec
            must be a preconfigured device.
    """
    if device_name:
        return preconfigured_device(device_name, strict=True)
    discovered_spec = discover_device_spec(content_type)
    configured_spec = preconfigured_device(discovered_spec.name, strict=False)
    return configured_spec or discovered_spec


def server_spec(content_type: str,
                device_name: Optional[str] = None) -> LslDataServer:
    """Get the device_spec definition to use for a mock data server. If the
    device_name is provided, the matching preconfigured_device will be used.
    Otherwise, the first matching preconfigured device with the given
    content_type will be used.

    Parameters
    ----------
        content_type - LSL content type (EEG, Gaze, etc).
        device_name - optional; name of the device. If provided, the DeviceSpec
            must be a preconfigured device.
    """
    if device_name:
        return preconfigured_device(device_name, strict=True)
    devices = with_content_type(content_type)
    if not devices:
        raise Exception(
            f"No configured devices have content_type {content_type}.")
    return devices[0]


def parse_stream_type(stream_type: str,
                      delimiter: str = "/") -> Tuple[str, str]:
    """Parses the stream type into a tuple of (content_type, device_name).

    Parameters
    ----------
        stream_type - LSL content type (EEG, Gaze, etc). If you know the name
            of the preconfigured device it can be added 'EEG/DSI-24'

    >>> parse_stream_type('EEG/DSI-24')
    ('EEG', 'DSI-24')
    >>> parse_stream_type('Gaze')
    ('Gaze', None)
    """
    if delimiter in stream_type:
        content_type, device_name = stream_type.split(delimiter)[0:2]
        return (content_type, device_name)
    return (stream_type, None)


def init_lsl_client(parameters: dict,
                    device_spec: DeviceSpec,
                    save_folder: str,
                    raw_data_file_name: str = None):
    """Initialize a client that acquires data from LabStreamingLayer."""

    data_buffer_seconds = round(max_inquiry_duration(parameters))

    return LslAcquisitionClient(max_buffer_len=data_buffer_seconds,
                                device_spec=device_spec,
                                save_directory=save_folder,
                                raw_data_file_name=raw_data_file_name)


def max_inquiry_duration(parameters: dict) -> float:
    """Computes the maximum duration of an inquiry based on the configured
    parameters. Paradigms which don't use all of the settings may have shorter
    durations. This can be used to determine the size of the data buffer.

    Parameters
    ----------
    - parameters : dict
        configuration details regarding the task and other relevant information.
         {
            "time_fixation": float,
            "time_prompt": float,
            "prestim_length": float,
            "stim_length": float,
            "stim_jitter": float,
            "task_buffer_length": float,
            "time_flash": float,

         }

    Returns
    -------
    duration in seconds
    """
    fixation_duration = parameters['time_fixation']
    target_duration = parameters['time_prompt']
    prestimulus_duration = parameters['prestim_length']
    poststim_duration = parameters['prestim_length']
    stim_count = parameters['stim_length']
    stim_duration = parameters['time_flash']
    interval_duration = parameters['task_buffer_length']
    jitter = parameters['stim_jitter']

    return prestimulus_duration + jitter + target_duration + fixation_duration + (
        stim_count * stim_duration) + poststim_duration + interval_duration


def analysis_channels(channels: List[str], device_spec: DeviceSpec) -> list:
    """Analysis Channels.

    Defines the channels within a device that should be used for analysis.

    Parameters
    ----------
    - channels(list(str)): list of channel names from the raw_data
    (excluding the timestamp)
    - device_spec(str): device from which the data was collected

    Returns
    --------
    A binary list indicating which channels should be used for analysis.
    If i'th element is 0, i'th channel in filtered_eeg is removed.
    """

    relevant_channels = device_spec.analysis_channels
    if not relevant_channels:
        raise Exception("Analysis channels for the given device not found: "
                        f"{device_spec.name}.")
    if channels is None:
        return relevant_channels
    return [int(ch in relevant_channels) for ch in channels]


def analysis_channel_names_by_pos(channels: List[str],
                                  channel_map: List[int]) -> Dict[int, str]:
    """Generate a dict mapping the index of a channel (used in offline
    analysis) to its name.

    Parameters:
    -----------
    - channels : list of channel names
    - channel_map : output of analysis_channels method; binary list
    where each item is either 0 or 1.
    """
    selection = [bool(x) for x in channel_map]
    selected_channels = np.array(channels)[selection]
    return {i: ch for i, ch in enumerate(selected_channels)}


def start_viewer(display_screen: int, parameter_location: str) -> None:
    """Start the data viewer application.

    Parameters
    ----------
    - display_screen : which monitor to use for display; usually 0 or 1.
    - parameter_location : path to parameters.json config file.
    """
    viewer = f'{BCIPY_ROOT}/gui/viewer/data_viewer.py'
    cmd = f'python {viewer} -m {display_screen} -p {parameter_location}'
    subprocess.Popen(cmd, shell=True)

    # hack: wait for window to open, so it doesn't error out when the main
    # window is open fullscreen.
    time.sleep(2)


def stream_types(acq_mode: str, delimiter: str = "+") -> List[str]:
    """Parse the provided acquisition mode into a list of LSL stream types.

    The list of supported/recommended types is available in the XDF wiki:
    https://github.com/sccn/xdf/wiki/Meta-Data

    However, some LSL driver apps deviate from this list so this function does
    not validate that the provided modes are in this list.

    Parameters
    ----------
        acq_mode - delimited list of stream types (ex. 'EEG+Gaze')
        delimiter - optional delimiter; default is '+'
    """
    return list(
        dict.fromkeys([mode.strip() for mode in acq_mode.split(delimiter)]))



import logging
import random
from typing import Any, List, Tuple, Union

import numpy as np
from psychopy import core, event, visual

from bcipy.config import SESSION_COMPLETE_MESSAGE
from bcipy.helpers.clock import Clock
from bcipy.helpers.stimuli import get_fixation
from bcipy.task.exceptions import InsufficientDataException

log = logging.getLogger(__name__)


def fake_copy_phrase_decision(copy_phrase, target_letter, text_task):
    """Fake Copy Phrase Decision.

    Parameters
    ----------
        copy_phrase(str): phrase to be copied
        target_letter(str): the letter supposed to be typed
        text_task(str): phrase spelled at this time

    Returns
    -------
        (next_target_letter, text_task, run) tuple
    """
    if text_task == '*':
        length_of_spelled_letters = 0
    else:
        length_of_spelled_letters = len(text_task)

    length_of_phrase = len(copy_phrase)

    if length_of_spelled_letters == 0:
        text_task = copy_phrase[length_of_spelled_letters]
    else:
        text_task += copy_phrase[length_of_spelled_letters]

    length_of_spelled_letters += 1

    # If there is still text to be spelled, update the text_task
    # and target letter
    if length_of_spelled_letters < length_of_phrase:
        next_target_letter = copy_phrase[length_of_spelled_letters]

        run = True

    # else, end the run
    else:
        run = False
        next_target_letter = None
        text_task = copy_phrase

    return next_target_letter, text_task, run


def calculate_stimulation_freq(flash_time: float) -> float:
    """Calculate Stimulation Frequency.

    In an RSVP paradigm, the inquiry itself will produce an
        SSVEP response to the stimulation. Here we calculate
        what that frequency should be based on the presentation
        time.

    PARAMETERS
    ----------
    :param: flash_time: time in seconds to present RSVP inquiry letters
    :returns: frequency: stimulation frequency of the inquiry
    """

    # We want to know how many stimuli will present in a second
    return 1 / flash_time


def construct_triggers(inquiry_timing: List[List]) -> List[Tuple[str, float]]:
    """Construct triggers from inquiry_timing data.

    Parameters
    ----------
    - inquiry_timing: list of tuples containing stimulus timing and text

    Returns
    -------
    list of (stim, offset) tuples, where offset is calculated relative to the
    first stim time.
    """
    if inquiry_timing:
        _, first_stim_time = inquiry_timing[0]
        return [(stim, ((timing) - first_stim_time))
                for stim, timing in inquiry_timing]
    return []


def target_info(triggers: List[Tuple[str, float]],
                target_letter: str = None,
                is_txt: bool = True) -> List[str]:
    """Targetness for each item in triggers.

    Parameters
    ----------
    - triggers : list of (stim, offset)
    - target_letter : letter the user is attempting to spell
    - is_txt : bool indicating whether the triggers are text stimuli

    Returns
    -------
    list of ('target' | 'nontarget') for each trigger.
    """
    fixation = get_fixation(is_txt)
    labels = {target_letter: 'target', fixation: 'fixation'}
    return [labels.get(trg[0], 'nontarget') for trg in triggers]


def get_data_for_decision(inquiry_timing,
                          daq,
                          offset=0.0,
                          prestim=0.0,
                          poststim=0.0):
    """Queries the acquisition client for a slice of data and processes the
    resulting raw data into a form that can be passed to signal processing and
    classifiers.

    Parameters
    ----------
    - inquiry_timing(list): list of tuples containing stimuli timing and labels. We assume the list progresses in
    - daq (DataAcquisitionClient): bcipy data acquisition client
    - offset (float): offset present in the system which should be accounted for when creating data for classification.
        This is determined experimentally.
    - prestim (float): length of data needed before the first sample to reshape and apply transformations
    - poststim (float): length of data needed after the last sample in order to reshape and apply transformations

    Returns
    -------
    (raw_data, triggers) tuple
    """
    _, first_stim_time = inquiry_timing[0]
    _, last_stim_time = inquiry_timing[-1]

    # adjust for offsets
    time1 = first_stim_time + offset - prestim
    time2 = last_stim_time + offset

    if time2 < time1:
        raise InsufficientDataException(
            f'Invalid data query [{time1}-{time2}] with parameters:'
            f'[inquiry={inquiry_timing}, offset={offset}, prestim={prestim}, poststim={poststim}]')

    # Construct triggers to send off for processing. This should not be zero anymore. it would be for prestim_len = 0
    triggers = [(text, ((timing - first_stim_time) + prestim))
                for text, timing in inquiry_timing]

    # Define the amount of data required for any processing to occur.
    data_limit = round((time2 - time1 + poststim) * daq.device_spec.sample_rate)
    log.debug(f'Need {data_limit} records for processing')

    # Query for raw data
    raw_data = daq.get_data(start=time1, limit=data_limit)

    if len(raw_data) < data_limit:
        message = f'Process Data Error: Not enough data received to process. ' \
            f'Data Limit = {data_limit}. Data received = {len(raw_data)}'
        log.error(message)
        raise InsufficientDataException(message)

    # Take only the sensor data from raw data and transpose it
    raw_data = np.array([
        np.array([_float_val(col) for col in record.data])
        for record in raw_data
    ],
        dtype=np.float64).transpose()

    return raw_data, triggers


def _float_val(col: Any) -> float:
    """Convert marker data to float values so we can put them in a
    typed np.array. The marker column has type float if it has a 0.0
    value, and would only have type str for a marker value."""
    if isinstance(col, str):
        return 1.0
    return float(col)


def trial_complete_message(win, parameters) -> List[visual.TextStim]:
    """Trial Complete Message.

    Function return a TextStim Object (see Psychopy) to complete the trial.

    Parameters
    ----------

        win (object): Psychopy Window Object, should be the same as the one
            used in the experiment
        parameters (dict): Dictionary of session parameters

    Returns
    -------
        array of message_stim (trial complete message to be displayed).
    """
    message_stim = visual.TextStim(
        win=win,
        height=parameters['info_height'],
        text=SESSION_COMPLETE_MESSAGE,
        font=parameters['font'],
        pos=(parameters['info_pos_x'],
             parameters['info_pos_y']),
        wrapWidth=None,
        color=parameters['info_color'],
        colorSpace='rgb',
        opacity=1, depth=-6.0)
    return [message_stim]


def print_message(window: visual.Window, message: str = "Initializing..."):
    """Draws a message on the display window using default config.

    Parameters
    ----------
        window (object): Psychopy Window Object, should be the same as the one
            used in the experiment
        parameters (dict): Dictionary of session parameters

    Returns
    -------
        TextStim object
    """
    message_stim = visual.TextStim(win=window, text=message)
    message_stim.draw()
    window.flip()
    return message_stim


def get_user_input(window, message, color, first_run=False):
    """Get User Input.

    Function returns whether or not to stop a trial. If a key of interest is
        passed (e.g. space), it will act on it.

    Parameters
    ----------

        window[psychopy task window]: task window.  *assumes wait_screen method

    Returns
    -------
        True/False: whether or not to stop a trial (based on escape key).
    """
    if not first_run:
        pause = False
        # check user input to make sure we should be going
        keys = event.getKeys(keyList=['space', 'escape'])

        if keys:
            # pause?
            if keys[0] == 'space':
                pause = True

            # escape?
            if keys[0] == 'escape':
                return False

    else:
        pause = True

    while pause:
        window.wait_screen(message, color)
        keys = event.getKeys(keyList=['space', 'escape'])

        if keys:
            if keys[0] == 'escape':
                return False
            pause = False

    return True


def get_key_press(
        key_list: List[str],
        clock: Clock,
        stamp_label: str = 'bcipy_key_press') -> Union[list, None]:
    """Get Key Press.

    A method to retrieve keys pressed of interest and get back a timestamp with
        a custom label


    Parameters
    ----------
        key_list(List[str]): list of keys to look for being pressed. Ex. ['space']
        clock(Clock): clock to use for timestamping any key press
        stamp_label(str): custom label to use for timstamping along with the key itself

    Returns
    -------
        Key Press Timing(List[stamp_label, timestamp])
    """
    response = event.getKeys(keyList=key_list, timeStamped=True)
    if response:
        # The timestamp from the response uses the psychopy.core.monotonicClock
        # which records the number of seconds since the experiment start (core
        # was imported).
        key, stamp = response[0]
        offset = clock.getTime() - core.getTime()
        timestamp = stamp + offset
        return [f'{stamp_label}_{key}', timestamp]
    return None


def pause_calibration(window, display, current_index: int, parameters: dict):
    """Pause calibration.

    Pauses calibration for a given number of seconds and displays a countdown
    to the user.


    PARAMETERS
    ----------
    :param: window: Currently active PsychoPy window
    :param: display: The current display
    :param: current_index: number of trials that have already taken place
    :param: trials_before_break: number of trials before break
    :param: break_len: length of the break time (in seconds)
    :param: break_message: message to display to the user during the break

    :returns: bool: break has taken place
    """
    # Check whether or not to present a break
    trials_before_break = parameters['trials_before_break']
    break_len = parameters['break_len']
    break_message = parameters['break_message']

    if (current_index != 0) and (current_index % trials_before_break) == 0:

        # present break message for break length
        for counter in range(break_len):
            time = break_len - counter
            message = f'{break_message} {time}s'
            display.update_task_state(
                text=message,
                color_list=['white'])
            display.draw_static()
            window.flip()
            core.wait(1)
        return True

    return False


def generate_targets(alp, stim_number):
    """Generate a list of targets for each trial, minimizing duplication."""
    if stim_number <= len(alp):
        return random.sample(alp, stim_number)

    # minimize duplicates
    times, remainder = divmod(stim_number, len(alp))

    lists = [random.sample(alp, len(alp)) for _ in range(times)]
    lists.append(random.sample(alp, remainder))

    # flatten list of lists
    targets = [target for sublist in lists for target in sublist]

    return targets



"""Defines classes that can write markers to LabStreamingLayer StreamOutlet."""
import logging
from typing import Any

import pylsl

log = logging.getLogger(__name__)


class MarkerWriter():
    """Abstract base class for an object that can be used to handle stimulus
    markers.
    """

    def push_marker(self, marker: Any):
        """Push the given stimulus marker for processing.

        Parameters
        ----------
        - marker : any object that can be converted to a str
        """
        raise NotImplementedError()

    def cleanup(self):
        """Performs any necessary cleanup"""
        raise NotImplementedError()


class LslMarkerWriter(MarkerWriter):
    """Writes stimulus markers to a LabStreamingLayer StreamOutlet
    using pylsl. To consume this data, the client code would need to create a
    pylsl.StreamInlet. See https://github.com/sccn/labstreaminglayer/wiki.
    """

    def __init__(self,
                 stream_name: str = "BCI_Stimulus_Markers",
                 stream_id: str = "bci_stim_markers"):
        super(LslMarkerWriter, self).__init__()
        self.stream_name = stream_name
        markers_info = pylsl.StreamInfo(stream_name, "Markers", 1, 0, 'string',
                                        stream_id)
        self.markers_outlet = pylsl.StreamOutlet(markers_info)
        self.first_marker_stamp: float = None

    def push_marker(self, marker: Any):
        """Push the given stimulus marker for processing.

        Parameters
        ----------
        - marker : any object that can be converted to a str
        """
        stamp = pylsl.local_clock()
        log.debug(f'Pushing marker {str(marker)} at {stamp}')
        self.markers_outlet.push_sample([str(marker)], stamp)
        if not self.first_marker_stamp:
            self.first_marker_stamp = stamp

    def cleanup(self):
        """Cleans up the StreamOutlet."""
        del self.markers_outlet


class NullMarkerWriter(MarkerWriter):
    """MarkerWriter which doesn't write anything.

    A NullMarkerWriter can be passed in to the calling object in scenarios
    where marker handling occurs indirectly (ex. through a trigger box). By
    using a NullMarkerWriter rather than a None value, the calling
    object does not have to do additional null checks and a separation
    of concerns is maintained regarding how triggers are written for different
    devices.

    See the Null Object Design Pattern:
    https://en.wikipedia.org/wiki/Null_object_pattern
    """

    def push_marker(self, marker: Any):
        pass

    def cleanup(self):
        pass



import logging
from typing import Optional

import numpy as np
from scipy.stats import iqr
from sklearn.neighbors import KernelDensity


class KernelDensityEstimate:
    """Kernel density estimate using scikit learn.
    Attr:
        bandwidth(float): bandwidth of the kernel
        scores(np.array): Shape (num_items, 2) - ratio of classification scores from RDA; used to compute bandwidth
        kernel(string): element to form the actual fitted pdf.
    """

    def __init__(self, scores: Optional[np.array] = None, kernel="gaussian", num_cls=2):
        bandwidth = 1.0 if scores is None else self._compute_bandwidth(scores, scores.shape[0])
        self.logger = logging.getLogger(__name__)
        self.logger.info(f"KDE. bandwidth={bandwidth}, kernel={kernel}")
        self.num_cls = num_cls
        self.list_den_est = [KernelDensity(bandwidth=bandwidth, kernel=kernel) for _ in range(self.num_cls)]

    def _compute_bandwidth(self, scores: np.array, num_items: int):
        """Estimate bandwidth parameter using Silverman's rule of thumb.
        See https://en.wikipedia.org/wiki/Kernel_density_estimation#A_rule-of-thumb_bandwidth_estimator

        Args:
            scores (np.array): Shape (num_items, 2) - positive and negative class probabilities from RDA
            num_channels (int): number of channels in the original data

        Returns:
            float: rule-of-thumb bandwidth parameter for KDE
        """
        bandwidth = 0.9 * min(np.std(scores), iqr(scores) / 1.34) * np.power(num_items, -0.2)
        return bandwidth

    def fit(self, x, y):
        """Fits the kernel density estimates base on labelled data.

        Args:
            x(ndarray[float]): shape (N) data array
            y(ndarray[float]): shape (N) label array
            Where N and c denotes number of samples and classes
            respectively.
        """
        for i, c in enumerate(np.unique(y)):
            dat = x[y == c]
            # Reshape is required, otherwise there's ambiguity if it's one
            # sample with N dims or N samples with 1 dims
            dat = np.squeeze(dat)
            dat = np.expand_dims(dat, axis=1)
            self.list_den_est[i].fit(dat)

    def transform(self, x):
        """Calculates likelihood ods of given data.
        Args:
            x(ndarray[float]): N x 1 data array
            Where N and c denotes number of samples and classes
        Returns:
             val(ndarray[float]): N x c  log-likelihood array
         respectively."""

        # Calculate likelihoods for each density estimate
        val = []
        for i in range(self.num_cls):
            dat = np.squeeze(x)
            dat = np.expand_dims(dat, axis=1)
            val.append(self.list_den_est[i].score_samples(dat))

        return np.transpose(np.array(val))

    def fit_transform(self, x, y):
        self.fit(x, y)
        return self.transform(x)
